from dataclasses import dataclass
from typing import Union, Sequence, Optional

import cv2
import numpy as np
from astropy import units as u
from astropy.units import Quantity
from numpy.typing import ArrayLike

# A coordinate is a sequence of two floats with an optional unit attached
CoordinateType = Union[Sequence[float], np.ndarray, Quantity]

# A transform is a 2x2 array with optional units attached
TransformType = Union[np.ndarray, Quantity, Sequence[Sequence[float]]]

# An extent is like a coordinate type.
# If it has a single element, this element is automatically reproduced along all dimensions
ExtentType = Union[float, CoordinateType]


def unitless(data: ArrayLike) -> np.ndarray:
    """
    Converts an object to a numpy array.
    Especially useful to remove the scaling from a unitless Quantity.

    Args:
        data: The input data.
        If `data` is a Quantity, it is converted to a (unitless) numpy array.
        All other data types are just returned as is.

    Returns:
        ArrayLike: unitless numpy array, or the input data if it is not a Quantity.

    Raises:
        UnitConversionError: If the data is a Quantity with a unit

    Note:
        Do NOT use `np.array(data)` to convert a Quantity to a numpy array,
        because this will drop the unit prefix.
        For example, `np.array(1 * u.s / u.ms) == 1`.
        Whereas `unitless(1 * u.s / u.ms) == 1000` gives the correct answer.

    Usage:
    >>> data = np.array([1.0, 2.0, 3.0]) * u.m
    >>> unitless_data = unitless(data / u.mm)
    """
    if isinstance(data, Quantity):
        return data.to_value(u.dimensionless_unscaled)
    else:
        return np.array(data)


@dataclass
class Transform:
    """Represents a transformation from one 2-d coordinate system to another.

    Transform objects are used to specify any combination of shift, (anisotropic) scaling, rotation and shear.
    The transform matrix, as well as the source and destination origins, can be specified with astropy units attached.

    Note that a Transform object does not contain any information about the extent and pixel size
    of the source or destination image, it only specifies the coordinate transformation itself.
    When a Transform object is used to transform image data (see :func:`project` and :func:`place`),
    the pixel_size information should be specified separately.
    """

    """
    Args:
        transform:
            2x2 transformation matrix that transforms (y,x)
            coordinates in the source image to (y,x) coordinates in the destination image.
            This matrix may include astropy units,
            for instance when converting from normalized SLM coordinates to physical coordinates in micrometers.
            Note:
                When specified, the units of the transform must match the units of the pixel_size of the destination
                divided by the pixel size of the source.
        source_origin:
            (y,x) coordinate of the origin in the source image, relative to the center of the image.
            By default, the center of the source image is taken as the origin of the transform,
            meaning that that point is mapped onto the destination_origin.
            Note:
                When specified, the units of source_origin must match the units of the pixel_size of the source.
        destination_origin:
            (y,x) coordinate of the origin in the destination image, relative to the center of the image.
            By default, the center of the destination image is taken as the origin of the transform,
            meaning that the source_origin is mapped to that point.
            Note:
                the units of destination_origin must match the units of the pixel_size of the source.


    """

    def __init__(
        self,
        transform: Optional[TransformType] = None,
        source_origin: Optional[CoordinateType] = None,
        destination_origin: Optional[CoordinateType] = None,
    ):
        self.transform = Quantity(transform if transform is not None else np.eye(2))
        self.source_origin = Quantity(source_origin) if source_origin is not None else None
        self.destination_origin = Quantity(destination_origin) if destination_origin is not None else None

        if source_origin is not None:
            self.destination_unit(self.source_origin.unit)  # check if the units are compatible

    def destination_unit(self, src_unit: u.Unit) -> u.Unit:
        """Computes the unit of the output of the transformation, given the unit of the input.

        Raises:
            ValueError: If src_unit does not match the unit of the source_origin (if specified) or
                if dst_unit does not match the unit of the destination_origin (if specified).
        """
        if self.source_origin is not None and not self.source_origin.unit.is_equivalent(src_unit):
            raise ValueError("src_unit must match the units of source_origin.")

        dst_unit = (self.transform[0, 0] * src_unit).unit
        if self.destination_origin is not None and not self.destination_origin.unit.is_equivalent(dst_unit):
            raise ValueError("dst_unit must match the units of destination_origin.")

        return dst_unit

    def cv2_matrix(
        self,
        source_shape: Sequence[int],
        source_pixel_size: CoordinateType,
        destination_shape: Sequence[int],
        destination_pixel_size: CoordinateType,
    ) -> np.ndarray:
        """Returns the transformation matrix in the format used by cv2.warpAffine.

        This matrix is 2x3, with the last column corresponding to the translation vector.

        Note: OpenCV uses the (x,y) convention, so the matrix is transposed compared to the numpy matrix.
        Note: OpenCV uses the _center_ of the top-left corner as the origin, whereas OpenWFS uses the centerso the origin is corrected.
            of the image as the origin. The difference of half the image size minus half the
            pixel size is corrected in the transformation matrix.
        """
        if min(source_shape) < 1 or min(destination_shape) < 1:
            raise ValueError("Image size must be positive")

        source_origin = 0.5 * (np.array(source_shape) - 1.0) * source_pixel_size
        if self.source_origin is not None:
            source_origin += self.source_origin

        destination_origin = 0.5 * (np.array(destination_shape) - 1.0) * destination_pixel_size
        if self.destination_origin is not None:
            destination_origin += self.destination_origin

        centered_transform = Transform(
            transform=self.transform,
            source_origin=source_origin,
            destination_origin=destination_origin,
        )

        # then convert the transform to a matrix, using the specified pixel sizes
        transform_matrix = centered_transform.to_matrix(
            source_pixel_size=source_pixel_size,
            destination_pixel_size=destination_pixel_size,
        )

        # finally, convert the matrix to the format used by cv2.warpAffine by swapping x and y columns and rows
        transform_matrix = transform_matrix[[1, 0, 2], 0:3]
        transform_matrix = transform_matrix[0:2, [1, 0, 2]]  # discard last row
        return transform_matrix

    def to_matrix(self, source_pixel_size: CoordinateType, destination_pixel_size: CoordinateType) -> np.ndarray:
        """Returns a homogeneous transformation matrix that transforms (y,x,1) coordinates to (y', x', 1)."""
        matrix = np.eye(3)
        matrix[0:2, 0:2] = unitless(self.transform * source_pixel_size / destination_pixel_size)
        if self.destination_origin is not None:
            matrix[0:2, 2] = unitless(self.destination_origin / destination_pixel_size)
        if self.source_origin is not None:
            matrix[0:2, 2] -= unitless((self.transform @ self.source_origin) / destination_pixel_size)
        return matrix

    def opencl_matrix(self) -> np.ndarray:
        """Returns the transformation matrix (including translation) in the format used by OpenCL.

        The returned matrix is a 3x4 matrix that includes the transformation matrix and the translation vector.
        The last two columns are for padding only.
        """

        # Compute the homogeneous transform matrix, and transpose it.
        # Then, swap x and y rows and columns,
        # and flip the sign of the y-axis
        matrix = self.to_matrix((1.0, 1.0), (1.0, 1.0)).transpose()

        # convert to opencl format: swap x and y rows and columns,
        # and flip the sign of the y-axis output.
        transform = np.eye(3, 4, dtype="float32", order="C")
        transform[:, 0] = matrix[[1, 0, 2], 1]
        transform[:, 1] = -matrix[[1, 0, 2], 0]
        return transform

    @staticmethod
    def zoom(scale: float):
        """Returns a transform that just zooms in the image."""
        return Transform(transform=np.array(((scale, 0.0), (0.0, scale))))

    def __matmul__(self, other):
        """The matrix multiplication operator is used to compose transformations,
        and to apply a transformation to a vector."""
        if isinstance(other, Transform):
            return self.compose(other)
        else:
            return self.apply(other)

    def apply(self, vector: CoordinateType) -> CoordinateType:
        """Applies the transformation to a column vector.

        If `vector` is a 2-D array, applies the transformation to each column of `vector` individually.
        """
        if self.source_origin is not None:
            vector = vector - self.source_origin
        vector = self.transform @ vector
        if self.destination_origin is not None:
            vector = vector + self.destination_origin
        return vector

    def inverse(self):
        """Compute the inverse transformation,
        such that the composition of the transformation and its inverse is the identity.
        """

        # invert the transform matrix
        if self.transform is not None:
            transform = np.linalg.inv(self.transform)
        else:
            transform = None

        # swap source and destination origins
        return Transform(
            transform,
            source_origin=self.destination_origin,
            destination_origin=self.source_origin,
        )

    def compose(self, other: "Transform"):
        """Compose two transformations.

        Args:
            other (Transform): the transformation to apply first

        Returns:
            Transform: the composition of the two transformations
        """
        transform = self.transform @ other.transform
        destination_origin = (
            self.apply(other.destination_origin) if other.destination_origin is not None else self.destination_origin
        )
        return Transform(transform, other.source_origin, destination_origin)

    def _standard_input(self) -> Quantity:
        """Construct standard input points (1,0), (0,1) and (0,0) with the source unit of this transform."""
        return Quantity(((1.0, 0.0, 0.0), (0.0, 1.0, 0.0)), self.source_origin)

    @classmethod
    def identity(cls):
        return Transform()


def place(
    out_shape: tuple[int, ...],
    out_pixel_size: Quantity,
    source: np.ndarray,
    offset: Optional[Quantity] = None,
    out: Optional[np.ndarray] = None,
):
    """Takes a source array and places it in an otherwise empty array of specified shape and pixel size.

    The source array must have a pixel_size property (see set_pixel_size).
    It is scaled (using area interpolation) so that the pixel size matches that of the output.
    Then, the source array is placed into the output array at the position given by offset,
    where `offset=None` or `offset=(0.0, 0.0) * pixel_size` corresponds to centering the source image
    with respect to the output array.
    Parts of the source array that extend beyond the output are cropped, and parts of the array that are not
    covered by the source array are zero padded.

    Note: this function currently works for 2-D inputs only

    Args:

    """
    out_extent = out_pixel_size * np.array(out_shape)
    transform = Transform(destination_origin=offset)
    return project(source, out_extent=out_extent, out_shape=out_shape, transform=transform, out=out)


def project(
    source: np.ndarray,
    *,
    source_extent: Optional[ExtentType] = None,
    transform: Optional[Transform] = None,
    out: Optional[np.ndarray] = None,
    out_extent: Optional[ExtentType] = None,
    out_shape: Optional[tuple[int, ...]] = None,
) -> np.ndarray:
    """Projects the input image onto an array with specified shape and resolution.

    The input image is scaled so that the pixel sizes match those of the output,
    and cropped/zero-padded so that the data shape matches that of the output.
    Optionally, an additional :class:`~Transform` can be specified, e.g., to scale or translate the source image.

    Args:
        source: input image.
        source_extent: extent of the source image in some physical unit.
            If not given (``None``), the extent metadata of the input image is used.
            see :func:`~get_extent`.
        transform: optional transformed (rotate, translate, etc.)
            to appy to the source image before placing it in the output
        out: optional array where the output image is stored in.
        out_extent: extent of the output image in some physical unit.
            If not given, the extent metadata of the out image is used.
        out_shape: shape of the output image.
            This value is ignored if `out` is specified.

    Returns:
        np.ndarray: the projected image (`out` if specified, otherwise a new array)
    """
    transform = transform if transform is not None else Transform()
    if out is not None:
        if out_shape is not None and out_shape != out.shape:
            raise ValueError("out_shape and out.shape must match. Note that out_shape may be omitted")
        if out.dtype != source.dtype:
            raise ValueError("out and source must have the same dtype")
        out_shape = out.shape
        out_extent = out_extent or get_extent(out)
    if out_shape is None:
        raise ValueError("Either out_shape or out must be specified")
    if out_extent is None:
        raise ValueError("Either out_extent or the pixel_size metadata of out must be specified")
    source_extent = source_extent if source_extent is not None else get_extent(source)
    source_ps = source_extent / np.array(source.shape)
    out_ps = out_extent / np.array(out_shape)

    t = transform.cv2_matrix(source.shape, source_ps, out_shape, out_ps)
    # swap x and y in matrix and size, since cv2 uses the (x,y) convention.
    out_size = (out_shape[1], out_shape[0])
    if (source.dtype == np.complex128) or (source.dtype == np.complex64):
        if out is None:
            out = np.zeros(out_shape, dtype=source.dtype)
        # real part
        out.real = cv2.warpAffine(
            source.real,
            t,
            out_size,
            flags=cv2.INTER_NEAREST,
            borderMode=cv2.BORDER_CONSTANT,
            borderValue=(0.0,),
        )
        # imaginary part
        out.imag = cv2.warpAffine(
            source.imag,
            t,
            out_size,
            flags=cv2.INTER_NEAREST,
            borderMode=cv2.BORDER_CONSTANT,
            borderValue=(0.0,),
        )

    else:
        dst = cv2.warpAffine(
            source,
            t,
            out_size,
            dst=out,
            flags=cv2.INTER_NEAREST,
            borderMode=cv2.BORDER_CONSTANT,
            borderValue=(0.0,),
        )
        if out is not None and out is not dst:
            raise ValueError("OpenCV did not use the specified output array. This should not happen.")
        out = dst
    return set_pixel_size(out, out_ps)


def set_pixel_size(data: ArrayLike, pixel_size: Optional[Quantity]) -> np.ndarray:
    """
    Sets the pixel size metadata for the given data array.

    Args:
        data (ArrayLike): The input data array.
        pixel_size (Optional[Quantity]): The pixel size to be set. When a single-element pixel size is given,
            it is broadcasted to all dimensions of the data array.
            Passing None sets the pixel size metadata to None.

    Returns:
        np.ndarray: The modified data array with the pixel size metadata.

    Usage:
    >>> data = np.array([[1, 2], [3, 4]])
    >>> pixel_size = 0.1 * u.m
    >>> modified_data = set_pixel_size(data, pixel_size)
    """
    data = np.array(data)

    if pixel_size is not None and pixel_size.size == 1:
        pixel_size = pixel_size * np.ones(data.ndim)

    data.dtype = np.dtype(data.dtype, metadata={"pixel_size": pixel_size})
    return data


def get_pixel_size(data: np.ndarray) -> Optional[Quantity]:
    """
    Extracts the pixel size metadata from the data array.

    Args:
        data (np.ndarray): The input data array or Quantity.

    Returns:
        OptionalQuantity]: The pixel size metadata, or None if no pixel size metadata is present.

    Usage:
    >>> import astropy.units as u
    >>> import numpy as np
    >>> data = set_pixel_size(((1, 2), (3, 4)), 5 * u.um)
    >>> pixel_size = get_pixel_size(data)

    TODO: return 1 if no pixel size is present
    """
    metadata = data.dtype.metadata
    if metadata is None:
        return None
    return data.dtype.metadata.get("pixel_size", None)


def get_extent(data: np.ndarray) -> Quantity:
    """
    Extracts the extent from the data array.
    The extent is always equal to `shape * pixel_size`.
    """
    pixel_size = get_pixel_size(data)
    if pixel_size is None:
        return Quantity(data.shape)
    return data.shape * pixel_size


def set_extent(data: np.ndarray, extent: ExtentType) -> np.ndarray:
    """
    Sets the extent metadata for the given data array.

    Args:
        data (ArrayLike): The input data array.
        extent (ArrayLike): The extent to be set. When a single-element extent is given,
            it is broadcasted to all dimensions of the data array.
            Passing None sets the extent metadata to None.

    Returns:
        np.ndarray: The modified data array with the extent metadata.

    Usage:
    >>> data = np.array([[1, 2], [3, 4]])
    >>> extent = 0.1 * u.m
    >>> modified_data = set_extent(data, extent)
    """
    return set_pixel_size(data, extent / np.array(data.shape))



import OpenGL.GL as GL
import numpy as np

from .context import Context


class Texture:
    def __init__(self, slm, texture_type=None):
        self.context = Context(slm)
        self.handle = GL.glGenTextures(1)
        self.type = texture_type if texture_type is not None else GL.GL_TEXTURE_2D
        self.synchronized = False  # self.data is not yet synchronized with texture in GPU memory
        self._data_shape = None  # current size of the texture, to see if we need to make a new texture or
        # overwrite the exiting one

        # create a single pixel texture as default (also activates the OpenGL context and binds the texture
        self.set_data(0)

        # set wrapping and interpolation options
        GL.glTexParameteri(self.type, GL.GL_TEXTURE_WRAP_S, GL.GL_REPEAT)
        GL.glTexParameteri(self.type, GL.GL_TEXTURE_WRAP_T, GL.GL_REPEAT)
        GL.glTexParameteri(self.type, GL.GL_TEXTURE_WRAP_R, GL.GL_REPEAT)
        GL.glTexParameteri(self.type, GL.GL_TEXTURE_MAG_FILTER, GL.GL_NEAREST)
        GL.glTexParameteri(self.type, GL.GL_TEXTURE_MIN_FILTER, GL.GL_NEAREST)

    def __del__(self):
        with self.context as slm:
            if slm:
                GL.glDeleteTextures(1, [self.handle])

    def _bind(self, idx):
        """Bind texture to texture unit idx. Assumes that the OpenGL context is already active."""
        GL.glActiveTexture(GL.GL_TEXTURE0 + idx)
        GL.glBindTexture(self.type, self.handle)

    def set_data(self, value):
        """Set texture data.

        The texture data is directly copied to the GPU memory,
         so the original data array can be modified or deleted.
        """
        value = np.asarray(value, dtype=np.float32, order="C")

        with self.context:
            GL.glBindTexture(self.type, self.handle)
            GL.glPixelStorei(GL.GL_UNPACK_ALIGNMENT, 4)  # alignment is at least four bytes since we use float32
            (internal_format, data_format, data_type) = (
                GL.GL_R32F,
                GL.GL_RED,
                GL.GL_FLOAT,
            )

            if self.type == GL.GL_TEXTURE_1D:
                # check if data has the correct dimension, convert scalars to arrays of correct dimension
                if value.ndim == 0:
                    value = value.reshape((1,))
                elif value.ndim != 1:
                    raise ValueError("Data should be a 1-d array or a scalar")
                if value.shape != self._data_shape:
                    # create a new texture
                    GL.glTexImage1D(
                        GL.GL_TEXTURE_1D,
                        0,
                        internal_format,
                        value.shape[0],
                        0,
                        data_format,
                        data_type,
                        value,
                    )
                    self._data_shape = value.shape
                else:
                    # overwrite existing texture
                    GL.glTexSubImage1D(
                        GL.GL_TEXTURE_1D,
                        0,
                        0,
                        value.shape[0],
                        data_format,
                        data_type,
                        value,
                    )

            elif self.type == GL.GL_TEXTURE_2D:
                if value.ndim == 0:
                    value = value.reshape((1, 1))
                elif value.ndim != 2:
                    raise ValueError("Data should be a 2-D array or a scalar")
                if value.shape != self._data_shape:
                    GL.glTexImage2D(
                        GL.GL_TEXTURE_2D,
                        0,
                        internal_format,
                        value.shape[1],
                        value.shape[0],
                        0,
                        data_format,
                        data_type,
                        value,
                    )
                    self._data_shape = value.shape
                else:
                    GL.glTexSubImage2D(
                        GL.GL_TEXTURE_2D,
                        0,
                        0,
                        0,
                        value.shape[1],
                        value.shape[0],
                        data_format,
                        data_type,
                        value,
                    )
            else:
                raise ValueError("Texture type not supported")

    def get_data(self):
        with self.context:
            data = np.empty(self._data_shape, dtype="float32")
            GL.glGetTextureImage(self.handle, 0, GL.GL_RED, GL.GL_FLOAT, data.size * 4, data)
            return data



import logging
import time

import astropy.units as u
import numpy as np
import pytest

from openwfs.processors import CropProcessor
from openwfs.simulation import StaticSource, NoiseSource, SLM
from openwfs.utilities import set_pixel_size, get_pixel_size


def test_set_pixel_size():
    # Test case 1: Broadcast pixel size
    data = np.array([[1, 2], [3, 4]])
    pixel_size = 0.1 * u.m
    modified_data = set_pixel_size(data, pixel_size)
    assert np.all(get_pixel_size(modified_data) == (0.1, 0.1) * u.m)

    # Test case 2: Anisotropic pixel size
    data = np.array([[1, 2], [3, 4]])
    pixel_size = [0.1, 0.2] * u.m
    modified_data = set_pixel_size(data, pixel_size)
    assert np.all(get_pixel_size(modified_data) == pixel_size)

    # Test case 3: None pixel size
    data = np.array([[1, 2], [3, 4]])
    pixel_size = None
    modified_data = set_pixel_size(data, pixel_size)
    assert np.all(get_pixel_size(modified_data) == pixel_size)

    # Test case 4: Getting pixel size from bare numpy array
    data = np.array([[1, 2], [3, 4]])
    assert get_pixel_size(data) is None


@pytest.mark.parametrize("pixel_size", [4 * u.um, (4, 3) * u.um])
def test_mock_detector(pixel_size):
    image = np.ones((4, 5))
    source = StaticSource(image, pixel_size=pixel_size)
    data = source.read()
    data2 = source.trigger().result()
    data3 = np.empty(data.shape)
    source.trigger(out=data3)
    source.wait()
    assert np.allclose(image, data)
    assert np.allclose(image, data2)
    assert np.allclose(image, data3)
    ps = get_pixel_size(data)
    assert len(ps) == 2
    assert len(source.pixel_size)
    assert np.all(ps == source.pixel_size)
    assert np.all(ps == pixel_size)
    y = source.coordinates(0)
    x = source.coordinates(1)
    assert np.allclose(y, ([0.5, 1.5, 2.5, 3.5] * ps[0]).reshape((4, 1)))
    assert np.allclose(x, ([0.5, 1.5, 2.5, 3.5, 4.5] * ps[1]).reshape((1, 5)))
    assert np.allclose(source.extent, (4, 5) * ps)


@pytest.mark.parametrize("duration", [0.0 * u.s, 0.5 * u.s])
def test_timing_detector(caplog, duration):
    # note: this test will fail if multithreading is disabled
    caplog.set_level(logging.DEBUG)
    image0 = np.zeros((4, 5))
    image1 = np.ones((4, 5))
    source = StaticSource(image0, pixel_size=4 * u.um, duration=duration)
    t0 = time.time_ns()
    f0 = source.trigger()
    t1 = time.time_ns()
    source.data = image1  # waits for data acquisition to complete
    t2 = time.time_ns()
    f1 = source.trigger()
    t3 = time.time_ns()
    assert np.allclose(f1.result(), image1)
    t4 = time.time_ns()
    assert np.allclose(f0.result(), image0)
    t5 = time.time_ns()

    assert np.allclose(t1 - t0, 0.0, atol=0.1e9)
    assert np.allclose(t2 - t1, duration.to_value(u.ns), atol=0.1e9)
    assert np.allclose(t3 - t2, 0.0, atol=0.1e9)
    assert np.allclose(t4 - t3, duration.to_value(u.ns), atol=0.1e9)
    assert np.allclose(t5 - t4, 0.0, atol=0.1e9)


def test_noise_detector():
    source = NoiseSource("uniform", data_shape=(10, 11, 20), low=-1.0, high=1.0, pixel_size=4 * u.um)
    data = source.read()
    assert data.shape == (10, 11, 20)
    assert np.min(data) >= -1.0
    assert np.max(data) < 1.0
    assert np.allclose(np.mean(data), 0.0, atol=0.1)
    assert np.allclose(np.std(data), 2.0 / np.sqrt(12.0), atol=0.1)
    assert np.all(get_pixel_size(data) == 4 * u.um)
    source.data_shape = (2, 3)
    assert source.read().shape == (2, 3)


def test_mock_slm():
    slm = SLM((4, 4))
    slm.set_phases(0.5)
    assert np.allclose(slm.pixels.read(), round(0.5 * 256 / (2 * np.pi)), atol=0.5 / 256)
    discretized_phase = slm.phases.read()
    assert np.allclose(discretized_phase, 0.5, atol=1.1 * np.pi / 256)
    assert np.allclose(slm.field.read(), np.exp(1j * discretized_phase[0, 0]), rtol=2 / 256)
    slm.set_phases(np.array(((0.1, 0.2), (0.3, 0.4))), update=False)
    assert np.allclose(slm.phases.read(), 0.5, atol=1.1 * np.pi / 256)  # slm.update() not yet called, so should be 0.5
    slm.update()
    assert np.allclose(
        slm.phases.read(),
        np.array(
            (
                (0.1, 0.1, 0.2, 0.2),
                (0.1, 0.1, 0.2, 0.2),
                (0.3, 0.3, 0.4, 0.4),
                (0.3, 0.3, 0.4, 0.4),
            )
        ),
        atol=1.1 * np.pi / 256,
    )


def test_crop():
    data = np.random.uniform(size=(4, 5))
    source = StaticSource(data, pixel_size=1 * u.um)
    cropped = CropProcessor(source, padding_value=np.nan)
    assert cropped.data_shape == (4, 5)
    c = cropped.read()
    assert np.all(c == data)

    cropped.pos = (1, 2)
    assert cropped.data_shape == (4, 5)
    c2 = cropped.read()
    assert c2.shape == cropped.data_shape
    assert np.all(c2[0:-1, 0:-2] == data[1:, 2:])
    assert np.all(np.isnan(c2[-1, :]))
    assert np.all(np.isnan(c2[:, -2:]))

    cropped.pos = (-1, -2)
    cropped.data_shape = (2, 4)
    c3 = cropped.read()
    assert c3.shape == (2, 4)
    assert np.all(np.isnan(c3[0, :]))
    assert np.all(np.isnan(c3[0:1, :]))
    assert np.all(c3[1:, 2:] == data[0:1, 0:2])


def test_crop_1d():
    data = np.random.uniform(size=(10,))
    source = StaticSource(data, pixel_size=1 * u.um)
    cropped = CropProcessor(source, padding_value=np.nan)
    assert cropped.data_shape == (10,)
    c = cropped.read()
    assert np.all(c == data)

    cropped.pos = 4
    assert cropped.data_shape == (10,)
    assert cropped.pos == (4,)
    c2 = cropped.read()
    assert c2.shape == cropped.data_shape
    assert np.all(c2[:-4] == data[4:])
    assert np.all(np.isnan(c2[-4:]))

    cropped.pos = 4
    cropped.data_shape = 2
    assert cropped.data_shape == (2,)
    c3 = cropped.read()
    assert c3.shape == cropped.data_shape
    assert np.all(c3 == data[4:6])


# TODO: translate the tests below.
#  They should test the SingleROI processor, checking if the returned averaged value is correct.
#
# def test_square_selection_detector():
#     """
#     Test if the square selection detector is working as expected.
#     """
#     sim = SimulatedWFS()
#     sim.set_ideal_wf(np.zeros([500, 500]))  # correct wf = flat
#
#     width = 20
#     height = 20
#     detector = SingleRoiSquare(sim, width=width, height=height, top=240, left=240)
#     detector.trigger()  #
#
#     # Assert condition for detector functionality
#     assert sim.read()[250, 250] / (width * height) == detector.read(), "Square detector not working as expected"
#
#
# def square_selection_detector_test():
#     sim = SimulatedWFS()
#     sim.set_ideal_wf(np.zeros([500, 500]))  # correct wf = flat
#
#     width = 20
#     height = 20
#     detector = SingleRoiSquare(sim, width=width, height=height, top=240, left=240)
#     detector.trigger()  #
#     if sim.read()[250, 250] / (width * height) != detector.read():
#         raise Exception(f"Square detector not working as expected")
#     return True
#
#
# @pytest.fixture
# def mock_simulated_wfs():
#     sim = SimulatedWFS()
#     sim.read = lambda: data.camera()  # Mock the read function to return a fixed image
#     return sim
#
#
# def test_select_roi_circle(mock_simulated_wfs):
#     """
#     Test the SelectRoiCircle detector functionality.
#     """
#     detector = SelectRoiCircle(mock_simulated_wfs)
#     detector.trigger()
#
#     circle_image = detector.read_circle()
#
#     assert circle_image is not None
#     assert circle_image.shape == mock_simulated_wfs.read().shape, "Circle ROI read image shape mismatch"
#
#
# def test_select_roi_square(mock_simulated_wfs):
#     """
#     Test the SelectRoiSquare detector functionality.
#     """
#     detector = SelectRoi(mock_simulated_wfs)
#     detector.trigger()
#
#     square_image = detector.read_square()
#
#     # Check the properties of square_image
#     assert square_image is not None



from typing import Optional

import astropy.units as u
import numpy as np
import pytest
import skimage
from scipy.linalg import hadamard
from scipy.ndimage import zoom

from . import complex_random
from openwfs.algorithms import (
    StepwiseSequential,
    FourierDualReference,
    DualReference,
    SimpleGenetic,
)
from openwfs.algorithms.troubleshoot import field_correlation
from openwfs.algorithms.utilities import WFSController
from openwfs.processors import SingleRoi
from openwfs.simulation import SimulatedWFS, StaticSource, SLM, Microscope
from openwfs.simulation.mockdevices import GaussianNoise, Camera


@pytest.fixture(autouse=True)
def reset():
    np.random.seed(42)  # for reproducibility


@pytest.mark.parametrize("algorithm", ["ssa", "fourier"])
@pytest.mark.parametrize("shape, feedback_shape", [((4, 7), (210,)), ((10, 7), (21, 10)), ((20, 31), (3, 7, 10))])
@pytest.mark.parametrize("noise", [0.0, 0.1])
def test_multi_target_algorithms(shape: tuple[int, int], feedback_shape: tuple[int, ...], noise: float, algorithm: str):
    """
    Test the multi-target capable algorithms (SSA and Fourier dual ref).

    This tests checks if the algorithm achieves the theoretical enhancement,
    and it also verifies that the enhancement and noise fidelity
    are estimated correctly by the algorithm.
    """
    M = np.prod(feedback_shape)  # number of targets
    phase_steps = 6

    # create feedback object, with noise if needed
    sim = SimulatedWFS(t=complex_random((*feedback_shape, *shape)))
    sim.slm.set_phases(0.0)
    I_0 = np.mean(sim.read())
    feedback = GaussianNoise(sim, std=I_0 * noise)

    if algorithm == "ssa":
        alg = StepwiseSequential(
            feedback=feedback,
            slm=sim.slm,
            n_x=shape[1],
            n_y=shape[0],
            phase_steps=phase_steps,
        )
        N = np.prod(shape)  # number of input modes
        # SSA is inaccurate if N is slightly lower because the reference beam varies with each segment.
        # The variation is of order 1/N in the signal, so the fidelity is (N-1)/N.
        # todo: check!
        alg_fidelity = (N - 1) / N
        signal = (N - 1) / N**2  # for estimating SNR
        masks = [True]  # use all segments for determining fidelity
    else:  # 'fourier':
        alg = FourierDualReference(
            feedback=feedback,
            slm=sim.slm,
            slm_shape=shape,
            k_radius=(np.min(shape) - 1) // 2,
            phase_steps=phase_steps,
        )
        N = len(alg.phase_patterns[0]) + len(alg.phase_patterns[1])  # number of input modes
        alg_fidelity = 1.0  # Fourier is accurate for any N
        signal = 1 / 2  # for estimating SNR.
        masks = alg.masks  # use separate halves of the segments for determining fidelity

    # Execute the algorithm to get the optimized wavefront
    # for all targets simultaneously
    result = alg.execute()

    # Determine the optimized intensities in each of the targets individually
    # Also estimate the fidelity of the transmission matrix reconstruction
    # This fidelity is determined row by row, since we need to compensate
    # the unknown phases. The normalization of the correlation function
    # is performed on all rows together, not per row, to increase
    # the accuracy of the estimate.
    # For the dual reference algorithm, the left and right half will have a different overall amplitude factor.
    # This is corrected for by computing the correlations for the left and right half separately
    #
    I_opt = np.zeros(feedback_shape)
    for b in np.ndindex(feedback_shape):
        sim.slm.set_phases(-np.angle(result.t[b]))
        I_opt[b] = feedback.read()[b]

    t_correlation = t_fidelity(result.t, sim.t, masks=masks)

    # Check the enhancement, noise fidelity and
    # the fidelity of the transmission matrix reconstruction
    coverage = N / np.prod(shape)
    print(signal)
    print(noise)
    theoretical_noise_fidelity = signal * phase_steps / (signal * phase_steps + noise**2)

    enhancement = I_opt.mean() / I_0
    theoretical_enhancement = np.pi / 4 * theoretical_noise_fidelity * alg_fidelity * (N - 1) + 1
    estimated_enhancement = result.estimated_enhancement.mean() * alg_fidelity
    theoretical_t_correlation = theoretical_noise_fidelity * alg_fidelity * coverage
    estimated_t_correlation = result.fidelity_noise * result.fidelity_calibration * alg_fidelity * coverage
    tolerance = 4.0 / np.sqrt(M)  # TODO: find out if this should be stricter
    print(
        f"\nenhancement:      \ttheoretical= {theoretical_enhancement},\testimated={estimated_enhancement},\tactual: {enhancement}"
    )
    print(
        f"t-matrix fidelity:\ttheoretical = {theoretical_t_correlation},\testimated = {estimated_t_correlation},\tactual = {t_correlation}"
    )
    print(f"noise fidelity:   \ttheoretical = {theoretical_noise_fidelity},\testimated = {result.fidelity_noise}")
    print(f"comparing at relative tolerance: {tolerance}")

    assert np.allclose(
        enhancement, theoretical_enhancement, rtol=tolerance
    ), f"""
        The SSA algorithm did not enhance the focus as much as expected.
        Theoretical {theoretical_enhancement}, got {enhancement}"""

    assert np.allclose(
        estimated_enhancement, enhancement, rtol=tolerance
    ), f"""
         The SSA algorithm did not estimate the enhancement correctly.
         Estimated {estimated_enhancement}, got {enhancement}"""

    assert np.allclose(
        t_correlation, theoretical_t_correlation, rtol=tolerance
    ), f"""
        The SSA algorithm did not measure the transmission matrix correctly.
        Expected {theoretical_t_correlation}, got {t_correlation}"""

    assert np.allclose(
        estimated_t_correlation, theoretical_t_correlation, rtol=tolerance
    ), f"""
        The SSA algorithm did not estimate the fidelity of the transmission matrix correctly.
        Expected {theoretical_t_correlation}, got {estimated_t_correlation}"""

    assert np.allclose(
        result.fidelity_noise, theoretical_noise_fidelity, rtol=tolerance
    ), f"""
        The SSA algorithm did not estimate the noise correctly.
        Expected {theoretical_noise_fidelity}, got {result.fidelity_noise}"""


def half_mask(shape):
    """
    Args:
        shape: shape of the output array

    Returns:
        Returns a boolean mask with [:, :shape[2]] set to False
        and [:, shape[1]] set to True].
    """
    mask = np.zeros(shape, dtype=bool)
    mask[:, shape[1] // 2 :] = True
    return mask


def t_fidelity(
    t: np.ndarray, t_correct: np.ndarray, *, columns: int = 2, masks: Optional[tuple[np.ndarray, ...]] = (True,)
) -> float:
    """
    Compute the fidelity of the measured transmission matrix.

    Since the overall phase for each row is unknown, the fidelity is computed row by row.
    Moreover, for dual-reference algorithms the left and right half of the wavefront
    may have different overall amplitude factors. This is corrected for by computing
    the fidelity for the left and right half separately.

    The fidelities for all rows (and halves) are weighted by the 'intensity' in the correct transmission matrix.

    Args:
        t: The measured transmission matrix. 'rows' of t are the *last* index
        t_correct: The correct transmission matrix
        columns: The number of columns in the transmission matrix
        masks: Masks for the left and right half of the wavefront, or None to use the full wavefront
    """
    fidelity = 0.0
    norm = 0.0
    for r in np.ndindex(t.shape[:-columns]):  # each row
        for m in masks:
            rm = t[r][m]
            rm_c = t_correct[r][m]
            fidelity += abs(np.vdot(rm, rm_c) ** 2 / np.vdot(rm, rm))
            norm += np.vdot(rm_c, rm_c)

    return fidelity / norm


def test_fourier2():
    """Test the Fourier dual reference algorithm using WFSController."""
    slm_shape = (10, 10)
    aberrations = skimage.data.camera() * ((2 * np.pi) / 255.0)
    sim = SimulatedWFS(aberrations=aberrations)
    alg = WFSController(
        FourierDualReference, feedback=sim, slm=sim.slm, slm_shape=slm_shape, k_radius=3.5, phase_steps=3
    )

    # check if the attributes of the algorithm were passed through correctly
    assert alg.k_radius == 3.5
    alg.k_radius = 2.5
    assert alg.k_radius == 2.5
    before = sim.read()
    alg.wavefront = WFSController.State.OPTIMIZED  # this will trigger the algorithm to optimize the wavefront
    after = sim.read()
    alg.wavefront = WFSController.State.FLAT  # this set the wavefront back to flat
    before2 = sim.read()
    assert before == before2
    assert after / before > 3.0


@pytest.mark.skip("Not implemented")
def test_fourier_microscope():
    aberration_phase = skimage.data.camera() * ((2 * np.pi) / 255.0) + np.pi
    aberration = StaticSource(aberration_phase, pixel_size=2.0 / np.array(aberration_phase.shape))
    img = np.zeros((1000, 1000), dtype=np.int16)
    signal_location = (250, 250)
    img[signal_location] = 100
    slm_shape = (1000, 1000)

    src = StaticSource(img, pixel_size=400 * u.nm)
    slm = SLM(shape=(1000, 1000))
    sim = Microscope(
        source=src,
        incident_field=slm.field,
        magnification=1,
        numerical_aperture=1,
        aberrations=aberration,
        wavelength=800 * u.nm,
    )
    cam = Camera(sim, analog_max=100)
    roi_detector = SingleRoi(cam, pos=(250, 250))  # Only measure that specific point
    alg = FourierDualReference(feedback=roi_detector, slm=slm, slm_shape=slm_shape, k_radius=1.5, phase_steps=3)
    controller = WFSController(alg)
    controller.wavefront = WFSController.State.FLAT_WAVEFRONT
    before = roi_detector.read()
    controller.wavefront = WFSController.State.SHAPED_WAVEFRONT
    after = roi_detector.read()
    # imshow(controller._optimized_wavefront)
    print(after / before)
    scaled_aberration = zoom(aberration_phase, np.array(slm_shape) / aberration_phase.shape)
    assert_enhancement(slm, roi_detector, controller._result, np.exp(1j * scaled_aberration))


def test_fourier_correction_field():
    """
    Check the field correlation between set aberration and optimized wavefront of the Fourier-based algorithm.
    """
    aberrations = skimage.data.camera() * (2.0 * np.pi / 255.0)
    sim = SimulatedWFS(aberrations=aberrations)
    alg = FourierDualReference(
        feedback=sim,
        slm=sim.slm,
        slm_shape=np.shape(aberrations),
        k_radius=3.0,
        phase_steps=3,
    )
    t = alg.execute().t

    t_correct = np.exp(1j * aberrations)
    correlation = np.vdot(t, t_correct) / np.sqrt(np.vdot(t, t) * np.vdot(t_correct, t_correct))

    # TODO: integrate with other test cases, duplication
    assert abs(correlation) > 0.75


def test_phase_shift_correction():
    """
    Test the effect of shifting the found correction of the Fourier-based algorithm.
    Without the bug, a phase shift of the entire correction should not influence the measurement.
    TODO: move to test of SimulatedWFS, since it is not testing the WFS algorithm itself
    """
    aberrations = skimage.data.camera() * (2.0 * np.pi / 255.0)
    sim = SimulatedWFS(aberrations=aberrations)
    alg = FourierDualReference(
        feedback=sim,
        slm=sim.slm,
        slm_shape=np.shape(aberrations),
        k_radius=1.5,
        phase_steps=3,
    )
    t = alg.execute().t

    # compute the phase pattern to optimize the intensity in target 0
    optimised_wf = -np.angle(t)
    sim.slm.set_phases(0)

    optimised_wf -= 5
    signals = []

    for n in range(5):
        optimised_wf += 2
        sim.slm.set_phases(optimised_wf)
        signal = sim.read()
        signals.append(signal)

    assert (
        np.std(signals) / np.mean(signals) < 0.001
    ), f"""The response of SimulatedWFS is sensitive to a flat 
        phase shift. This is incorrect behaviour"""


@pytest.mark.parametrize("optimized_reference", [True, False])
@pytest.mark.parametrize("type", ["flat", "phase_step", "amplitude_step"])
def test_flat_wf_response_fourier(optimized_reference, type):
    """
    Test the response of the Fourier-based WFS method when the solution is flat
    A flat solution means that the optimal correction is no correction.
    Also tests if stitching is done correctly by having an aberration pattern which is flat (but different) on the two halves.

    test the optimized wavefront by checking if it has irregularities.
    """
    t = np.ones(shape=(4, 4), dtype=np.complex64)
    if type == "phase_step":
        t[:, 2:] = np.exp(2.0j)
    elif type == "amplitude_step":
        t[:, 2:] = 2.0

    sim = SimulatedWFS(t=t)

    alg = FourierDualReference(
        feedback=sim,
        slm=sim.slm,
        slm_shape=np.shape(t),
        k_radius=1.5,
        phase_steps=3,
        optimized_reference=optimized_reference,
    )

    result = alg.execute()
    assert (
        abs(field_correlation(result.t / np.abs(result.t), t / np.abs(t))) > 0.99
    ), "The phases were not calculated correctly"
    assert t_fidelity(result.t, t, masks=alg.masks) > 0.99, "The amplitudes were not calculated correctly"


def test_flat_wf_response_ssa():
    """
    Test the response of the SSA WFS method when the solution is flat.
    """
    aberrations = np.zeros(shape=(512, 512))
    sim = SimulatedWFS(aberrations=aberrations)

    alg = StepwiseSequential(feedback=sim, slm=sim.slm, n_x=4, n_y=4, phase_steps=3)

    # Execute the SSA algorithm to get the optimized wavefront
    t = alg.execute().t
    optimised_wf = np.angle(t)

    # Assert that the standard deviation of the optimized wavefront is below the threshold,
    # indicating that it is effectively flat
    assert np.std(optimised_wf) < 0.001, f"Response flat wavefront not flat, std: {np.std(optimised_wf)}"


def test_multidimensional_feedback_ssa():
    aberrations = np.random.uniform(0.0, 2 * np.pi, (5, 2, 256, 256))
    sim = SimulatedWFS(aberrations=aberrations)

    alg = StepwiseSequential(feedback=sim, slm=sim.slm)
    t = alg.execute().t

    # compute the phase pattern to optimize the intensity in target 2,1
    target = (2, 1)
    optimised_wf = -np.angle(t[(*target, ...)])

    # Calculate the enhancement factor
    # Note: technically this is not the enhancement, just the ratio after/before
    sim.slm.set_phases(0.0)
    before = sim.read()
    sim.slm.set_phases(optimised_wf)
    after = sim.read()
    enhancement = after / before

    assert (
        enhancement[target] >= 3.0
    ), f"""The SSA algorithm did not enhance focus as much as expected.
            Expected at least 3.0, got {enhancement}"""


def test_multidimensional_feedback_fourier():
    aberrations = np.random.uniform(0.0, 2 * np.pi, (5, 2, 256, 256))
    sim = SimulatedWFS(aberrations=aberrations)

    # input the camera as a feedback object, such that it is multidimensional
    alg = FourierDualReference(feedback=sim, slm=sim.slm, k_radius=3.5, phase_steps=3)
    t = alg.execute().t

    # compute the phase pattern to optimize the intensity in target 0
    optimised_wf = -np.angle(t[2, 1, :, :])

    # Calculate the enhancement factor
    # Note: technically this is not the enhancement, just the ratio after/before
    sim.slm.set_phases(0.0)
    before = sim.read()
    sim.slm.set_phases(optimised_wf)
    after = sim.read()
    enhancement = after / before

    assert (
        enhancement[2, 1] >= 3.0
    ), f"""The algorithm did not enhance the focus as much as expected.
            Expected at least 3.0, got {enhancement}"""


@pytest.mark.parametrize("population_size, elite_size", [(30, 15)])  # , (30, 5)])
def test_simple_genetic(population_size: int, elite_size: int):
    """
    Test the SimpleGenetic algorithm.
    Note: this is not very rigid test, as we currently don't have theoretical expectations for the performance.
    """
    shape = (100, 71)
    sim = SimulatedWFS(t=complex_random(shape), multi_threaded=False)
    alg = SimpleGenetic(
        feedback=sim,
        slm=sim.slm,
        shape=shape,
        population_size=population_size,
        elite_size=elite_size,
        generations=1000,
    )
    result = alg.execute()
    sim.slm.set_phases(0.0)
    before = sim.read()
    sim.slm.set_phases(-np.angle(result.t))
    after = sim.read()

    print(after / before)
    assert after / before > 4


@pytest.mark.parametrize("basis_str", ("plane_wave", "hadamard"))
@pytest.mark.parametrize("shape", ((8, 8), (16, 4)))
def test_dual_reference_ortho_split(basis_str: str, shape: tuple[int, int]):
    """Test dual reference in iterative mode with an orthonormal phase-only basis.
    Two types of bases are tested: plane waves and Hadamard"""
    N = shape[0] * (shape[1] // 2)
    modes_shape = (N, shape[0], shape[1] // 2)
    if basis_str == "plane_wave":
        # Create a full plane wave basis for one half of the SLM.
        phases = np.angle(np.fft.fft2(np.eye(N).reshape(modes_shape), axes=(1, 2)))
    elif basis_str == "hadamard":
        phases = np.angle(hadamard(N).reshape(modes_shape))
    else:
        raise f'Unknown type of basis "{basis_str}".'

    mask = half_mask(shape)
    phases_set = np.pad(phases, ((0, 0), (0, 0), (0, shape[1] // 2)))

    sim = SimulatedWFS(t=complex_random(shape))

    alg = DualReference(
        feedback=sim,
        slm=sim.slm,
        phase_patterns=(phases_set, np.flip(phases_set, axis=2)),
        group_mask=mask,
        iterations=4,
    )

    # Checks for orthonormal basis properties
    assert np.allclose(alg.gram[0], np.eye(N), atol=1e-6)  # Gram matrix must be I

    # Cobasis vectors are just the complex conjugates
    assert np.allclose(alg.cobasis[0], np.exp(-1j * phases_set) * abs(alg.cobasis[0]), atol=1e-6)

    # Test phase-only field correlation
    result = alg.execute()
    sim_t_phase_only = np.exp(1j * np.angle(sim.t))
    result_t_phase_only = np.exp(1j * np.angle(result.t))
    assert np.abs(field_correlation(sim_t_phase_only, result_t_phase_only)) > 0.999

    assert t_fidelity(result.t, sim.t, masks=alg.masks) > 0.9


def test_dual_reference_non_ortho_split():
    """
    Test dual reference with a non-orthogonal basis.
    """
    # Create set of modes that are barely linearly independent
    N1 = 6
    N2 = 3
    M = N1 * N2
    mode_set_half = np.exp(2j * np.pi / 3 * np.eye(M).reshape((N2, N1, M))).T / np.sqrt(M)

    # note: typically we just use the other half for the mode set B, but here we set the half to 0 to
    # make sure it is not used.
    mode_set = np.pad(mode_set_half, ((0, 0), (0, 0), (0, N2)))
    phases_set = np.angle(mode_set)
    mask = half_mask((N1, 2 * N2))

    # Create aberrations
    x = np.linspace(-1, 1, N1).reshape((1, -1))
    y = np.linspace(-1, 1, N1).reshape((-1, 1))
    aberrations = (np.sin(0.8 * np.pi * x) * np.cos(1.3 * np.pi * y) * (0.8 * np.pi + 0.4 * x + 0.4 * y)) % (2 * np.pi)
    aberrations[0:1, :] = 0
    aberrations[:, 0:2] = 0

    sim = SimulatedWFS(aberrations=aberrations)

    alg = DualReference(
        feedback=sim,
        slm=sim.slm,
        phase_patterns=(phases_set, np.flip(phases_set, axis=2)),
        group_mask=mask,
        phase_steps=4,
        iterations=4,
    )

    result = alg.execute()

    aberration_field = np.exp(1j * aberrations)
    t_field = np.exp(1j * np.angle(result.t))

    assert np.abs(field_correlation(aberration_field, t_field)) > 0.999



import threading
import time
import weakref

import astropy.units as u
import numpy as np
from astropy.units import Quantity
from serial.tools import list_ports
from zaber_motion import Units, ascii, binary

from ..core import Actuator


class _SerialPortConnection:
    """Wrapper for a serial port connection managed by the zaber_motion library.

    This wrapper automatically closes the connection when it is deleted.
    Note: this may not be needed if Zaber already does this automatically,
        but that behavior is not documented.
    Note: do not instantiate this class directly.
        Instead use `_SerialPortConnection.open` to open a connection,
        and re-use an existing connection if possible.
    """

    def __init__(self, port: str, protocol: str):
        self.protocol = protocol.lower()
        if self.protocol == "ascii":
            self.connection = ascii.Connection.open_serial_port(port)
        elif self.protocol == "binary":
            self.connection = binary.Connection.open_serial_port(port)
        else:
            raise ValueError("protocol must be 'ascii' or 'binary'")

    def __del__(self):
        self.connection.close()

    _ports = weakref.WeakValueDictionary()
    _lock = threading.RLock()

    @staticmethod
    def open(port: str, protocol: str | None = None) -> "_SerialPortConnection":
        """Opens a connection to a serial port.

        Establishes or retrieves an existing serial port connection, ensuring that only one connection
        is allowed per port with the same protocol. If a connection for the given port exists but uses
        a different protocol, an exception is raised.

        Args:
            port: The identifier of the serial port, e.g. "COM3"
            protocol ("ascii" | "binary" | None):
                The protocol associated with the connection.
                When 'None' and the port was already open, use the protocol with which it was opened earlier.
                If no port was open, use "ascii".

        Returns:
            _SerialPortConnection: A new or existing serial port connection for the given port.

        Raises:
            RuntimeError: If the specified port is already open with a different protocol.
        """
        with _SerialPortConnection._lock:
            connection = _SerialPortConnection._ports.get(port)
            if connection is not None:
                if protocol is not None and connection.protocol != protocol:
                    raise RuntimeError(f"Port {port} is already open with a different connection class.")
            else:
                connection = _SerialPortConnection(port, protocol or "ascii")
                _SerialPortConnection._ports[port] = connection

            return connection


class ZaberXYStage(Actuator):
    """
    Wrapper for a pair of Zaber linear stages connected via serial ports, controlling X and Y axes.

    todo: if we add more stages, it makes sense to make a general XYStage class that combines two linear stages into one.
    todo: Currently both axis move after each other, need to implement parallel movement.

    Args:
        port_x: COM port string for X axis, e.g. "COM3"
        port_y: COM port string for Y axis, e.g. "COM4". If None, uses port_x.
        device_number_x: index of the device on the X axis port (0-based)
        device_number_y: index of the device on the Y axis port (0-based). If None and port_y==port_x, assumes second device (1), else 0.
        protocol: "ascii" or "binary"

    Example:
        stage = ZaberXYStage(port_x="COM3", port_y="COM4", device_number_x=0, device_number_y=0, protocol="ascii")
        stage.home()  # home both axes
        stage.x = 5000 * u.um  # move X to 5000 micrometers
        stage.y = 2 * u.mm  # move Y to 2 millimeters
        print(stage.x, stage.y)  # get current positions
    """

    def __init__(
        self,
        port_x: str,
        port_y: str | None = None,
        device_number_x: int = 0,
        device_number_y: int | None = None,
        protocol: str = "ascii",
    ):
        # Initialize base class
        super().__init__(duration=np.inf * u.ms, latency=0 * u.ms)

        # If only one port is given, use it for both axes
        if port_y is None:
            port_y = port_x

        # If only one port is given and device_number_y is not specified, assume second device on same port and set device_number_y=1,
        # otherwise set device_number_y=0
        if device_number_y is None:
            device_number_y = 1 if port_y == port_x else 0

        # Create stage objects
        self.stage_x = ZaberLinearStage(port_x, device_number=device_number_x, protocol=protocol)
        self.stage_y = ZaberLinearStage(port_y, device_number=device_number_y, protocol=protocol)

    @property
    def x(self) -> Quantity[u.um]:
        return self.stage_x.position

    @x.setter
    def x(self, value: Quantity[u.um]):
        self.stage_x.position = value

    @property
    def y(self) -> Quantity[u.um]:
        return self.stage_y.position

    @y.setter
    def y(self, value: Quantity[u.um]):
        self.stage_y.position = value

    def home(self):
        self.stage_x.home()
        self.stage_y.home()

    def x_home(self):
        self.stage_x.home()

    def y_home(self):
        self.stage_y.home()

    @staticmethod
    def list_all_devices():
        """
        List all COM ports and devices connected to them using both ASCII and Binary protocols.
        Works even if some ports are already open by objects.

        Returns: dict { "COMx": [{"protocol": "ascii"|"binary", "devices": [...] }], ... }
        """
        return ZaberLinearStage.list_all_devices()


class ZaberLinearStage(Actuator):
    """
    Wrapper for a single Zaber linear stage connected via serial port.

    todo: asynchroneous behavior (now setting the position blocks until the stage has moved completely)
    todo: add a `settle_time` parameter to tell openwfs how long to wait after a move before starting a measurement.
    todo: at the moment it is possible to construct two stage objects referring to the same device. This condition
        should be checked and result in an error being raised.

    Args:
        port: COM port string, e.g. "COM3"
        device_number: index of the device on the port (0-based)
        protocol: "ascii" or "binary"

    Example:
        stage = ZaberLinearStage(port="COM3", device_number=0, protocol="ascii")
        stage.x = 5000 * u.um  # move to 5000 micrometers
        print(stage.x)  # get current position
        stage.home()  # home the stage
    """

    def __init__(self, port: str, device_number: int = 0, protocol: str = "ascii"):
        Actuator.__init__(self, duration=0 * u.ms, latency=0 * u.ms)

        # open a serial port connection and store the object to keep the connection alive
        self._serial_port = _SerialPortConnection.open(port, protocol)

        # store device number (for debugging only), and obtain an object for controlling the device
        self._device_number = device_number
        try:
            self._stage = self._serial_port.connection.detect_devices()[device_number]
        except IndexError:
            raise RuntimeError(
                f"No Zaber device found on port {port} at position {device_number} using {protocol} protocol."
            )

    @property
    def position(self) -> Quantity[u.um]:
        return self._stage.get_position(unit=Units.LENGTH_MICROMETRES) * u.um

    @position.setter
    def position(self, value: Quantity[u.um]):
        super()._start()  # wait for all previous commands to finish
        # Note: at the moment setting the stage position blocks until the stage has
        # moved completely. This should be changed so that 'busy' checks
        # if the movement has finished, but this behavior is not easily supported by
        # the binary Zaber API, which is needed for older stages.
        try:
            self._stage.move_absolute(value.to_value(u.um), Units.LENGTH_MICROMETRES)
        finally:
            # Set the end time to *now*. This tells the framework that it can start using
            # cameras and other detectors.
            # todo: here we should add a stabilization time to take into account
            #   vibrations after the motor stops.
            self._end_time_ns = time.time_ns()

    def home(self):
        self._stage.home()

    @staticmethod
    def list_all_devices():
        """
        List all COM ports and devices connected to them using both ASCII and Binary protocols.
        Works even if some ports are already open by objects.

        Returns: dict { "COMx": [{"protocol": "ascii"|"binary", "devices": [...] }], ... }
        """
        all_devices = {}
        for port in list_ports.comports():
            for protocol in ["ascii", "binary"]:
                try:
                    serial_port = _SerialPortConnection.open(port.device, protocol)
                except RuntimeError:
                    continue  # happens if the port is currently open using a different protocol.
                print(f"Port: {port}")
                try:
                    devices = serial_port.connection.detect_devices()
                    all_devices[port] = {"protocol": protocol, "devices": devices}
                    print(f"  Protocol: {protocol}")
                    for i, d in enumerate(devices):
                        print(f"    Device {i}: {d}")
                except Exception:
                    pass
                serial_port = None  # closes the port

        return all_devices



import time

import astropy.units as u
import nidaqmx as ni
from astropy.units import Quantity


class Gain:
    """
    A device that controls the voltage of a PMT gain using a NI data acquisition card.

    It is controlling a hamamatsu M9012 power supply, supplying power for a PMT H7422-40.

    It uses an analogue out connection to set the voltage,
    an analogue in channel to check the gain's self-protecting overload status,
    and a digital out that sends the signal to the gain to reset itself.

    It contains a boolean called reset that triggers the method on_reset.
    Ideally, we would like some button to execute methods in our devices,
    but this approach allows the user to execute methods in the device property manager.
    """

    def __init__(self, port_ao, port_ai, port_do, reset=False, gain=0 * u.V):
        """
        Initializes the Gain control.

        Args:
            port_ao (str): Port name for analog output (controls PMT voltage).
            port_ai (str): Port name for analog input (monitors overload status).
            port_do (str): Port name for digital output (controls gain reset).
            reset (bool): Initial state of the reset trigger.
            gain (Quantity[u.V]): Initial gain value.
        """
        self.port_ao = port_ao
        self.port_ai = port_ai
        self.port_do = port_do
        self._reset = reset
        self._gain = gain
        self.gain = gain  # triggers write to NI-DAQ

    def check_overload(self):
        with ni.Task() as task:
            task.ai_channels.add_ai_voltage_chan(self.port_ai)
            in_stream = task.in_stream

            data = in_stream.read(number_of_samples_per_channel=1)
            if data > 2.5:
                overload = True
            else:
                overload = False

            return overload

    def on_reset(self, value):
        if value:
            with ni.Task() as task:
                task.do_channels.add_do_chan(
                    self.port_do, line_grouping=nidaqmx.constants.LineGroupingLineGrouping.CHAN_FOR_ALL_LINES
                )
                task.write([True])
                time.sleep(1)
                task.write([False])

    @property
    def reset(self) -> bool:
        """Triggers the reset process if set to True."""
        return self._reset

    @reset.setter
    def reset(self, value: bool):
        self.on_reset(value)
        self._reset = value

    @property
    def gain(self) -> Quantity[u.V]:
        """Sets or gets the current gain value. Range: 0 to 0.9 volts."""
        # The range of values is the hardware supplier's defined voltage range. Setting the range here for safety
        return self._gain

    @gain.setter
    def gain(self, value: Quantity[u.V]):
        self._gain = value.to(u.V)
        with ni.Task() as write_task:
            channel = write_task.ao_channels.add_ao_voltage_chan(self.port_ao)
            channel.ao_min = 0
            channel.ao_max = 0.9
            write_task.write(self._gain.to_value(u.V))



from .basic_fourier import FourierDualReference
from .dual_reference import DualReference
from .genetic import SimpleGenetic
from .ssa import StepwiseSequential
from .troubleshoot import troubleshoot



import itertools

import numpy as np

from .utilities import WFSResult, DummyProgressBar
from ..core import Detector, PhaseSLM


class SimpleGenetic:
    """Simple genetic algorithm for wavefront shaping.

    This algorithm is included for illustrative purposes. It is based on the algorithm described
    in [1] and [2].

    The algorithm performs the following steps:

    1. Initialize all wavefronts in the population with random phases

    2. For each generation:

      2.1. Determine the feedback signal for each wavefront.

      2.2. Select the 'elite_size' best wavefronts to keep. Replace the rest with new wavefronts.
        2.2.1 If the elite wavefronts are too similar (> 97% identical elements),
              randomly generate the new wavefronts.

        2.2.2 Otherwise, generate new wavefronts by randomly selecting two elite wavefronts
              and mixing them randomly (element-wise).
              Then perform a mutation that replaces a fraction (`mutation_probability`) of the
              elements by a new value.

    References
    ----------
    [^1]: Conkey D B, Brown A N, Caravaca-Aguirre A M and Piestun R 'Genetic algorithm optimization
          for focusing through turbid media in noisy environments' Opt. Express 20 4840–9 (2012).
    [^2]: Benjamin R Anderson et al. 'A modular GUI-based program for genetic algorithm-based
          feedback-assisted wavefront shaping', J. Phys. Photonics 6 045008 (2024).
    """

    def __init__(
        self,
        feedback: Detector,
        slm: PhaseSLM,
        shape: tuple[int, int] = (500, 500),
        population_size: int = 30,
        elite_size: int = 5,
        generations: int = 100,
        mutation_probability: float = 0.005,
        generator=None,
    ):
        """
        Args:
            feedback: Source of feedback
            slm: The spatial light modulator
            shape: Width × height (in segments) of the wavefront
            population_size (int): The number of individuals in the population
            elite_size (int): The number of individuals in the elite pool
            generations (int): The number of generations
            mutation_probability (int): Fraction of elements in the offspring to mutate
            generator: a `np.random.Generator`, defaults to np.random.default_rng()

        """
        if np.prod(feedback.data_shape) != 1:
            raise ValueError("Only scalar feedback is supported")
        self.feedback = feedback
        self.slm = slm
        self.shape = shape
        self.population_size = population_size
        self.elite_size = elite_size
        self.generations = generations
        self.generator = generator or np.random.default_rng()
        self.mutation_count = round((population_size - elite_size) * np.prod(shape) * mutation_probability)

    def _generate_random_phases(self, shape):
        return self.generator.random(size=shape, dtype=np.float32) * (2 * np.pi)

    def execute(self, *, progress_bar=DummyProgressBar()) -> WFSResult:
        """Executes the algorithm.
        Args:
            progress_bar: Optional tqdm-like progress bar for displaying progress
        """

        # Initialize the population
        population = self._generate_random_phases((self.population_size, *self.shape))

        # initialize the progress bar if available
        progress_bar.total = self.generations * self.population_size

        for i in itertools.count():
            # Try all phase patterns
            measurements = np.zeros(self.population_size, dtype=np.float32)
            for p in range(self.population_size):
                self.slm.set_phases(population[p])
                self.feedback.trigger(out=measurements[p, ...])
                progress_bar.update()

            self.feedback.wait()

            # Sort the measurements in ascending order
            sorted_indices = np.argsort(measurements)
            elite = sorted_indices[-self.elite_size :]
            plebs = sorted_indices[: -self.elite_size]

            # Terminate after the specified number of generations, return the best wavefront
            if i >= self.generations:
                return WFSResult(t=np.exp(-1.0j * population[sorted_indices[-1]]), axis=2)

            # We keep the elite individuals, and regenerate the rest by mixing the elite
            # For this mixing, the probability of selecting an individual is proportional to its measured intensity.
            probabilities = measurements[elite]
            probabilities /= np.sum(probabilities)
            couples = self.generator.choice(elite, size=(2, len(plebs)), p=probabilities)

            if np.mean(population[couples[0]] == population[couples[1]]) > 0.97:
                # if the parents are too similar, randomly generate the plebs
                population[plebs] = self._generate_random_phases((len(plebs), *self.shape))
            else:
                # otherwise, mix and mutate the parents
                mix_masks = self.generator.integers(1, size=(len(plebs), *self.shape), dtype=bool)
                offspring = np.where(mix_masks, population[couples[0]], population[couples[1]])
                mutations = self.generator.integers(offspring.size, size=self.mutation_count)
                offspring.ravel()[mutations] = self._generate_random_phases(self.mutation_count)
                population[plebs] = offspring



from typing import Union, Sequence, Optional

import numpy as np
from astropy.units import Quantity

from .utilities import ExtentType, CoordinateType, unitless

# shape of a numpy array, or a single integer that is broadcast to a square shape
ShapeType = Union[int, Sequence[int]]

# a scalar quantity with optional unit attached
ScalarType = Union[float, np.ndarray, Quantity]

"""
Library of functions to create commonly used patterns

Each of the functions takes a `shape` input, which may be a scalar integer or a 2-element Sequence of integers
indicating the size (shape) in pixels of the returned field. If `shape` is a scalar, the same value is
used for both axes.

For the coordinates, the OpenGL convention is used, where the coordinates indicate the centers of the pixels.
By default, the returned pattern is assumed to cover a -1,1 x -1,1 square, 
which corresponds to a default `extent` parameter of (2.0, 2.0).
In this case, the coordinates range from -1+dx/2 to 1-dx/2, where dx=2.0/shape is the pixel size.

Exptent and shape may be specified individually to work with anisotropic pixels or rectangular patterns.
For example, a square pattern with anisotropic pixels may be described by shape=(80,100) and extent(2,2)
whereas shape=(80,100) and extent(8,10) describes square pixels that form a rectangle.

In a pupil-conjugate configuration, a disk of extent=(NA, NA) exactly covers the back pupil of the microscope objective.
The transformation matrix of the SLM should be set such that SLM coordinates correspond to normalized pupil coordinates.

The extent may have a unit of measure. In this case, other parameters (such as `radius`) may need to have
an according unit of measure.

The (0,0) coordinate is always located in the center of the pattern, which may be on a grid point (for odd shape)
or between grid points (for even shape).

The returned array has a pixel_size property attached.
"""


def coordinate_range(
    shape: ShapeType, extent: ExtentType, offset: Optional[CoordinateType] = None
) -> (Quantity, Quantity):
    """
    Returns coordinate vectors for the two coordinates (y and x).

    Given a range from `-extent/2` to `+extent/2`, uniformly divided into `shape` pixels,
    the returned coordinates correspond to the centers of these coordinates.

    Arguments:
        shape (ShapeType): size of the full grid (y, x) in pixels
        extent (ExtentType): extent of the coordinate range
        offset (Optional[CoordinateType]): offset to be added to the coordinates (optional)

    Returns:
        Tuple[Quantity, Quantity]: coordinate vectors for the two coordinates (y and x)
    """
    if np.size(shape) == 1:
        shape = np.array(shape).repeat(2)

    extent = Quantity(extent)
    if extent.size == 1:
        extent = extent.repeat(2)

    if offset is None:
        offset = 0.0 * extent  # by default, let center be (0,0) (in whatever unit)

    def c_range(res, ex, cx):
        dx = ex / res
        return np.arange(res) * dx + (0.5 * dx - 0.5 * ex + cx)

    return (
        c_range(shape[0], extent[0], offset[0]).reshape((-1, 1)),
        c_range(shape[1], extent[1], offset[1]).reshape((1, -1)),
    )


def r2_range(shape: ShapeType, extent: ExtentType):
    """Convenience function to return square distance to the origin
    Equivalent to computing cx^2 + cy^2
    """
    c0, c1 = coordinate_range(shape, extent)
    return c0**2 + c1**2


def tilt(
    shape: ShapeType,
    g: ExtentType,
    extent: ExtentType = (2.0, 2.0),
    phase_offset: float = 0.0,
):
    """Constructs a linear gradient pattern φ=2g·r

    Note, these are the Zernike tilt modes (modes 2 and 3 in the Noll index convention) with normalization
    so that :math:`\frac{\int_0^{2\pi} \int_0^1 |Z(\rho, \phi)|^2 \rho d\rho d\phi}{\int_0^{2\pi} \int_0^1 \rho d\rho d\phi} = 1`.

    Args:
        shape: see module documentation
        g(tuple of two floats): gradient vector.
          This has the unit: 1 / extent.unit.
          For the default extent of (2.0, 2.0), a value of g=(1,0)
          corresponds to having a ramp from -2 to +2 over the height of the pattern.
          With an extent of (2.0, 2.0) covering the full NA,
          this pattern causes a displacement of -2/π times the Abbe diffraction limit
          (Note: a positive x-gradient g causes the focal point to move in the _negative_ x-direction)
        extent: see module documentation
        phase_offset: optional additional phase offset to be added to the pattern
    """
    c0, c1 = coordinate_range(shape, extent * (Quantity(g) * 2.0))
    return unitless(c0 + (c1 + phase_offset))


def lens(shape: ShapeType, f: ScalarType, wavelength: ScalarType, extent: ExtentType):
    """Constructs a square texture that represents a wavefront defocus: (f-sqrt(f²+r²)) · 2π/λ

    `extent`, `wavelength` and `f` should have compatible units (typically astropy length units).

    Args:
        shape(ShapeType): see module documentation
        f(ScalarType): focal length
        wavelength(ScalarType): wavelength
        extent(ExtentType): physical extent of the SLM, same units as `f` and `wavelength`
    """
    r_sqr = r2_range(shape, extent)
    return unitless((f - np.sqrt(f**2 + r_sqr)) * (2 * np.pi / wavelength))


def propagation(
    shape: ShapeType,
    distance: ScalarType,
    wavelength: ScalarType,
    extent: ExtentType,
    refractive_index: ScalarType = 1.0,
):
    """Computes a wavefront that corresponds to digitally propagating the field in the object plane.

    k_z = sqrt(n² k_0²-k_x²-k_y²)
    φ = k_z · distance

    Args:
          shape: see module documentation
          distance (ScalarType): physical distance to propagate axially.
          refractive_index (Scalar):
          wavelength (Scalar):
            the numerical aperture, refractive index and wavelength are used
            to convert the `extent` from pupil coordinates to k-space (unit radians/meter),
          extent: extent of the returned image, in NA units. To cover the full NA with a square, use (2*NA, 2*NA)
    """
    # convert pupil coordinates to absolute k_x, k_y coordinates
    n_p2 = r2_range(shape, Quantity(extent))
    n_z = np.sqrt(np.maximum(refractive_index**2 - n_p2, 0.0))
    return unitless(2.0 * np.pi / wavelength * distance * n_z)


def disk(shape: ShapeType, radius: ScalarType = 1.0, extent: ExtentType = (2.0, 2.0)):
    """Constructs an image of a centered (ellipsoid) disk.

    (x / rx)^2 + (y / ry)^2 <= 1.0

    Args:
          shape: see module documentation
          radius (ScalarType): radius of the disk, should have the same unit as `extent`.
          extent: see module documentation
    """
    return 1.0 * (r2_range(shape, extent) < radius**2)


def gaussian(
    shape: ShapeType,
    waist: ScalarType,
    truncation_radius: ScalarType = None,
    extent: ExtentType = (2.0, 2.0),
):
    """Constructs an image of a centered Gaussian

    `waist`, `extent` and the optional `truncation_radius` should all have the same unit.

    Args:
        shape: see module documentation
        waist (ScalarType): location of the beam waist (1/e value)
            relative to half of the size of the pattern (i.e. relative to the `radius` of the square)
        truncation_radius (ScalarType): when not None, specifies the radius of a disk that is used to truncate the
            Gaussian. All values outside the disk are set to 0.
        extent: see module documentation

    """
    r_sqr = r2_range(shape, extent)
    w2inv = -1.0 / waist**2
    gauss = np.exp(unitless(r_sqr * w2inv))
    if truncation_radius is not None:
        gauss = gauss * disk(shape, truncation_radius, extent=extent)
    return unitless(gauss)



import pytest
import numpy as np
from openwfs.utilities.patterns import tilt


@pytest.mark.parametrize("shape", [10, (7, 10)])
def test_tilt(shape):
    t = tilt(shape, (5, 0.71))
    if np.size(shape) == 1:
        shape = (shape, shape)

    # for a default extent of 2.0, 2.0, the tilt should span 4 * t
    phase_diff0 = 5 * 4 * (1.0 - 1.0 / shape[0])
    phase_diff1 = 0.71 * 4 * (1.0 - 1.0 / shape[1])
    assert np.allclose(t[-1, 0] - t[0, 0], phase_diff0)
    assert np.allclose(t[-1, -1] - t[0, -1], phase_diff0)
    assert np.allclose(t[0, -1] - t[0, 0], phase_diff1)
    assert np.allclose(t[-1, -1] - t[-1, 0], phase_diff1)



import logging
import threading
import time
from abc import ABC, abstractmethod
from concurrent.futures import Future, ThreadPoolExecutor
from typing import Set, final, Tuple, Optional
from weakref import WeakSet

import astropy.units as u
import numpy as np
from astropy.units import Quantity
from numpy.typing import ArrayLike

from .utilities import set_pixel_size


class Device(ABC):
    """Base class for detectors and actuators.

    This is the abstract base class for all devices in the openwfs framework,
    including both detectors and actuators.

    See :ref:`key_concepts` for more information.
    """

    __slots__ = (
        "_end_time_ns",
        "_timeout_margin",
        "_locking_thread",
        "_error",
        "__weakref__",
        "_latency",
        "_duration",
        "_multi_threaded",
    )
    _workers = ThreadPoolExecutor(thread_name_prefix="Device._workers")
    _moving = False
    _state_lock = threading.Lock()
    _devices: "Set[Device]" = WeakSet()

    def __init__(self, *, duration: Quantity[u.ms], latency: Quantity[u.ms]):
        """Constructs a new Device object.

        Args:
            duration: Maximum amount of time it takes to perform the measurement or for the
                actuator to stabilize. This value does not include the latency.
                If the duration of an operation is not known in advance,
                (e.g., when waiting for a hardware trigger), value should be `np.inf * u.ms`.
                If the duration is variable, it is possible to override the `duration` property.
            latency: Minimum amount of time between sending a command or trigger to the device
                and the moment the device starts responding.
        """
        self._latency = latency
        self._duration = duration
        self._end_time_ns = 0
        self._timeout_margin = 5 * u.s
        self._locking_thread = None
        self._error = None
        Device._devices.add(self)

    @property
    @abstractmethod
    def _is_actuator(self):
        """Indicates whether the device is an actuator (True) or a detector (False)."""
        ...

    def _start(self):
        """Switches the state to 'moving' (for actuators) or 'measuring' (for detectors).

        This function changes the global state to 'moving' or 'measuring' if needed,
        and it may block until this state switch is completed.

        After switching, stores the time at which the operation will have ended in the ``_end_time_ns``
        field (i.e., ``time.time_ns() + self.latency + self.duration``).
        """

        # acquire a global lock, to prevent multiple threads to switch moving/measuring state simultaneously
        with Device._state_lock:
            # check if transition from moving/measuring or vice versa is needed
            if Device._moving != self._is_actuator():
                if Device._moving:
                    logging.debug("switch to MEASURING requested by %s.", self)
                else:
                    logging.debug("switch to MOVING requested by %s.", self)

                same_type = [device for device in Device._devices if device._is_actuator == self._is_actuator]
                other_type = [device for device in Device._devices if device._is_actuator != self._is_actuator]

                # compute the minimum latency of same_type
                # for instance, when switching to 'measuring', this number tells us how long it takes before any of the
                # detectors actually starts a measurement.
                # If this is a positive number, we can make the switch to 'measuring' slightly _before_
                # all actuators have stabilized.
                latency = min([device.latency for device in same_type], default=0.0 * u.ns)  # noqa - incorrect warning

                # wait until all devices of the other type have (almost) finished
                for device in other_type:
                    device.wait(up_to=latency)

                # changes the state from moving to measuring and vice versa
                Device._moving = not Device._moving
                if Device._moving:
                    logging.debug("state is now MOVING.")
                else:
                    logging.debug("state is now MEASURING.")

            # also store the time we expect the operation to finish
            # note: it may finish slightly earlier since (latency + duration) is a maximum value
            self._end_time_ns = time.time_ns() + self.latency.to_value(u.ns) + self.duration.to_value(u.ns)

    @property
    def latency(self) -> Quantity[u.ms]:
        """Minimum amount of time between sending a command or trigger to the device and the
        moment the device starts responding.

        The default value is 0.0 ms.

        Note:
            The latency is used to compute when it is safe to switch the global state from 'moving'
            to 'measuring' or vice versa.
            Devices that report a non-zero latency promise not to do anything before the latency has passed.
            This allows making the state switch even before the devices of the other type have all finished.

            This construct is used for spatial light modulators, which typically have a long latency (1-2 frames).
            Due to this latency, we may send an `update` command to the SLM even before the camera has finished reading
            the previous frame.

        Note:
            A device is allowed to report a different latency every time `latency` is called.
            For example, for a spatial light modulator that only refreshes
            at a fixed rate, we can add the remaining time until the next refresh to the latency.
        """
        return self._latency

    @property
    def duration(self) -> Quantity[u.ms]:
        """Maximum amount of time it takes to perform the measurement or for the
        actuator to stabilize.

        This value *does not* include the latency.

        For a detector, this is the maximum amount of time that elapses between returning from `trigger()` and the
        end of the measurement.
        For an actuator, this is the maximum amount of time that elapses from returning from a command like `update()`
        and the stabilization of the device.

        If the duration of an operation is not known in advance,
        (e.g., when waiting for a hardware trigger), this function should return `np.inf * u.ms`.

        Note:
            A device may update the duration dynamically.
            For example, a stage may compute the required time to
            move to the target position and update the duration accordingly.

        Note:
            If `latency` is a (lower) estimate, the duration should be high enough to guarantee that `latency +
            duration` is at least as large as the time between starting the operation and finishing it.
        """
        return self._duration

    def wait(self, up_to: Optional[Quantity[u.ms]] = None) -> None:
        """Waits until the device is (almost) in the `ready` state, i.e., has finished measuring or moving.

        This function is called by `_start` automatically to ensure proper synchronization between detectors and
        actuators, and it is called by `__del__` to ensure the device is not active when it is destroyed.
        The only time to call `wait` explicitly is when using pipelined measurements, see `Detector.trigger()`.

        For devices that report a duration (`duration ≠ ∞`), this function waits until
        `current_time - up_to >= self._end_time_ns`,
        where `_end_time_ns` was set by the last call to `_start`.

        For devices that report no duration `duration = ∞`, this function repeatedly calls `busy` until `busy`
        returns `False`. In this case, `up_to` is ignored.

        Args:
            up_to(Quantity[u.ms]):
                when specified, specifies that this function may return 'up_to' milliseconds
                *before* the device is finished.

        Raises:
            Any other exception raised by the device in another thread (e.g., during `_fetch`).
            TimeoutError: if the device has `duration = ∞`, and `busy` does not return `True` within
                `self.timeout`
            RuntimeError: if `wait` is called from inside a setter or from inside `_fetch`.
                This would cause a deadlock.

        """
        if self._error is not None:
            e = self._error
            self._error = None
            raise e

        # If duration = ∞, poll busy until it returns False or a timeout occurs.
        if not np.isfinite(self._end_time_ns):
            start = time.time_ns()
            timeout = self.timeout.to_value(u.ns)
            while self.busy():
                time.sleep(0.01)
                if time.time_ns() - start > timeout:
                    raise TimeoutError("Timeout in %s (tid %i)", self, threading.get_ident())
        else:
            time_to_wait = self._end_time_ns - time.time_ns()
            if up_to is not None:
                time_to_wait -= up_to.to_value(u.ns)
            if time_to_wait > 0:
                time.sleep(time_to_wait / 1.0e9)

    def busy(self) -> bool:
        """Returns true if the device is measuring or moving (see `wait()`).

        Note: if a device does not define a finite `duration`, it must override this function to poll for
        finalization."""
        return time.time_ns() < self._end_time_ns

    @property
    def timeout(self) -> Quantity[u.ms]:
        """Time after which a timeout error is raised when waiting for the device.

        The timeout is automatically adjusted if the `duration` changes.
        The default value is `duration + 5 s`."""
        duration = self.duration
        if np.isfinite(duration):
            return self._timeout_margin + duration
        else:
            return self._timeout_margin

    @timeout.setter
    def timeout(self, value):
        duration = self.duration
        if not np.isinf(duration):
            self._timeout_margin = value - duration
        else:
            self._timeout_margin = value.to(u.ms)


class Actuator(Device, ABC):
    """Base class for all actuators.

    Actuators are devices that can modify the state of the system, such as
    spatial light modulators, stages, or other control devices. This class
    extends the Device base class and sets the _is_actuator property to True.
    """

    __slots__ = ()

    @final
    def _is_actuator(self):
        return True


class Detector(Device, ABC):
    """Base class for all detectors, cameras and other data sources with possible dynamic behavior.

    Detectors are devices that can measure or acquire data from the system, such as
    cameras, sensors, or other measurement devices. This class extends the Device
    base class and sets the _is_actuator property to False.

    See :ref:`Detectors` in the documentation for more information.
    """

    __slots__ = (
        "_measurements_pending",
        "_lock_condition",
        "_pixel_size",
        "_data_shape",
    )

    def __init__(
        self,
        *,
        data_shape: Optional[tuple[int, ...]],
        pixel_size: Optional[Quantity],
        duration: Optional[Quantity[u.ms]],
        latency: Optional[Quantity[u.ms]],
        multi_threaded: bool = True,
    ):
        """
        Constructor for the Detector class.

        Args:
            data_shape:  The shape of the data array that `read()` will return. When None is passed,
                the subclass should override the `data_shape` property to return the actual shape.
            pixel_size:  The pixel size (in astropy length units). None if the pixels do not have a size.
                Subclassed can override the `pixel_size` property to return the actual pixel size.
            duration: The maximum amount of time that elapses between returning from `trigger()`
                and the end of the measurement. If the duration of an operation is not known in advance,
                (e.g., when waiting for a hardware trigger), this value should be `np.inf * u.ms`
                and the `busy` method should be overridden to return `False` when the measurement is finished.
                If None is passed, the subclass should override the `duration` property to return the actual duration.
            latency: The minimum amount of time between sending a command or trigger to the device
                and the moment the device starts responding. If None is passed, the subclass should override
                the `latency` property to return the actual latency.
            multi_threaded: If True, `_fetch` is called from a worker thread. Otherwise, `_fetch` is called
                directly from `trigger`. If the device is not thread-safe, or threading provides no benefit,
                or for easy debugging, set this to False.
        """
        super().__init__(duration=duration, latency=latency)
        self._measurements_pending = 0
        self._lock_condition = threading.Condition()
        self._error = None
        self._data_shape = data_shape
        self._pixel_size = pixel_size
        self._multi_threaded = multi_threaded

    @final
    def _is_actuator(self):
        return False

    def _increase_measurements_pending(self):
        with self._lock_condition:
            self._measurements_pending += 1

    def _decrease_measurements_pending(self):
        with self._lock_condition:
            self._measurements_pending -= 1
            if self._measurements_pending == 0:
                self._lock_condition.notify_all()

    def wait(self, up_to: Quantity[u.ms] = None) -> None:
        """Waits until the hardware has (almost) finished measuring

        Due to the automatic synchronization between detectors and actuators, this function only needs to be called
        explicitly when waiting for data to be stored in the `out` argument of :meth:`~.Detector.trigger()`.

        Args:
            up_to: if specified, this function may return `up_to` milliseconds *before* the hardware
                has finished measurements.
                If None, this function waits until the hardware has finished all measurements *and* all data is fetched,
                and stored in the `out` array if that was passed to trigger().

        """
        super().wait(up_to)
        if up_to is None:
            # wait until all pending measurements are processed.
            with self._lock_condition:
                while self._measurements_pending > 0:
                    self._lock_condition.wait()

    def trigger(self, *args, out=None, immediate=False, **kwargs) -> Future:
        """Triggers the detector to start acquisition of the data.

        This function does not wait for the measurement to complete.
        Instead, it returns a ``concurrent.futures.Future``.
        Call ``.result()`` on the returned object to wait for the data.
        Here is a typical usage pattern:

        .. code-block:: python

            # Trigger the detector, which starts the data capture process
            future = detector.trigger()

            # Do some other work, perhaps trigger other detectors to capture
            # data simultaneously...

            # Now read the data from the detector. If the data is not ready yet,
            # this will block until it is.
            data = future.result()

        An alternative method for asynchronous data capture is to use
        the `out` parameter to specify a location where to store the data:

        .. code-block:: python

            out = np.zeros((2,), dtype='float32')
            detector.trigger(out=out[0])  # start the first measurement
            detector.trigger(out=out[1])  # queue the second measurement
            detector.wait()  # wait for both measurements to complete
            # Now the data is stored in the `out` array.

        All input parameters are passed to the _fetch function of the detector.
        Child classes may override trigger() to call `super().trigger()` with additional parameters.
        If any of these parameters is a Future, it is awaited before calling _fetch.
        This way, data from multiple sources can be combined (see Processor).

        Note:
            To implement hardware triggering, do not override this function.
            Instead, override `_do_trigger()` to ensure proper synchronization and locking.

        Args:
            out: If specified, the data is stored in this array once it is available.
            immediate: If True, the data is fetched in the current thread. This is useful for debugging,
                and for cases where the data is needed immediately. It avoids the overhead (and debugging complications)
                of dispatching the call to _fetch to a worker thread.
            *args: Additional arguments passed to the _fetch function.
                If any of these arguments is a `concurrent.futures.Future`, the data is awaited before calling _fetch,
                and the data is passed instead of the `Future`.
                This is useful for combining data from multiple sources (see `Processor`).
            **kwargs: Additional keyword arguments passed to the _fetch function. Any `concurrent.futures.Future`
                in the keyword arguments is awaited before calling _fetch,
                and the data is passed instead of the `Future`.
        """
        self._increase_measurements_pending()
        try:
            self._start()
            self._do_trigger()
        except:  # noqa  - ok, we are not really catching the exception, just making sure the lock gets released
            self._decrease_measurements_pending()
            raise

        logging.debug("triggering %s (tid: %i).", self, threading.get_ident())
        if immediate or not self._multi_threaded:
            result = Future()
            result.set_result(self.__do_fetch(out, *args, **kwargs))  # noqa
            return result
        else:
            return Device._workers.submit(self.__do_fetch, out, *args, **kwargs)

    def __do_fetch(self, out_, *args_, **kwargs_):
        """Helper function that awaits all futures in the keyword argument list, and then calls _fetch"""
        try:
            if len(args_) > 0 or len(kwargs_) > 0:
                logging.debug("awaiting inputs for %s (tid: %i).", self, threading.get_ident())
            awaited_args = [(arg.result() if isinstance(arg, Future) else arg) for arg in args_]
            awaited_kwargs = {key: (arg.result() if isinstance(arg, Future) else arg) for (key, arg) in kwargs_.items()}
            logging.debug("fetching data of %s ((tid: %i)).", self, threading.get_ident())
            data = self._fetch(*awaited_args, **awaited_kwargs)
            data = set_pixel_size(data, self.pixel_size)
            assert data.shape == self.data_shape
            if out_ is not None:
                out_[...] = data  # store data in the location specified during trigger
            return data
        except Exception as e:
            # if we are storing the result in an `out` array,
            # the user may never call result() on the returned future object,
            # and the error may be lost.
            # Therefore, store it so that it can be raised on the next call to wait()
            if out_ is not None:
                self._error = e
            raise e  # raise the error again, it will be stored in the 'Future' object that was returned by trigger()
        finally:
            self._decrease_measurements_pending()

    def __new__(cls, *args, **kwargs):
        """This method is called before __init__ to create a new instance of the class.

        We need to override this to add attributes that will be used in __setattr__
        """
        instance = super().__new__(cls)
        instance._multi_threaded = False
        return instance

    def __setattr__(self, key, value):
        """Prevents modification of public attributes and properties while the device is locked.

        For detectors, this prevents modification of the detector settings while a measurement is in progress.
        For all devices, it prevents concurrent modification in a multi-threading context.

        Private attributes can be set without locking.
        Note that this is not thread-safe and should be done with care!!
        """

        # note: the check needs to be in this order, otherwise we cannot initialize set _multi_threaded
        if not key.startswith("_") and self._multi_threaded:
            with self._lock_condition:
                while self._measurements_pending > 0:
                    self._lock_condition.wait()
            super().__setattr__(key, value)
        else:
            super().__setattr__(key, value)

    def _do_trigger(self) -> None:
        """Override this function to perform the actual hardware trigger."""
        pass

    @abstractmethod
    def _fetch(self, *args, **kwargs) -> np.ndarray:
        """Read the data from the detector

        Args:
            The args and kwargs are passed from the call to trigger()
        Note:
            After reading the data, the Detector attaches the pixel_size metadata.
        Note:
            Child classes must implement this function, and store the data in `out[...]` if `out` is not None.
        """
        ...

    @final
    def read(self, *args, immediate=True, **kwargs) -> np.ndarray:
        """Triggers the detector and waits for the data to arrive.

        Shortcut for trigger().result().
        """
        return self.trigger(*args, immediate=immediate, **kwargs).result()

    @property
    def data_shape(self) -> Tuple[int, ...]:
        """The shape of the data array that `read()` will return.

        For some detectors this property may be mutable, for example, for a camera it
        represents the height and width of the ROI, which can be changed.
        """
        return self._data_shape

    @property
    def pixel_size(self) -> Optional[Quantity]:
        """Physical dimension of one element in the returned data array.

        For cameras, this is the pixel size (in astropy length units).
        For detectors returning a time trace, this value is specified in astropy time units.
        The pixel_size is a 1-D array with an element for each dimension of the returned data.

        By default, the pixel size cannot be set.
        However, in some cases (such as when the `pixel_size` is actually a sampling interval),
        it makes sense for the child class to implement a setter.
        """
        return self._pixel_size

    @final
    def coordinates(self, dimension: int) -> Quantity:
        """Returns an array with the coordinate values along the d-th axis.

        The coordinates represent the _centers_ of the grid points. For example,
        for an array of shape ``(2,)`` the coordinates are `[0.5, 1.5] * pixel_size`
        and not `[0, 1] * pixel_size`. If `self.pixel_size is None`, a pixel size
        of 1.0 is used.

        The coordinates are returned as an array with the same number of
        dimensions as `data_shape`, with the d-th dimension holding the coordinates.
        This facilitates meshgrid-like computations, e.g.
        `cam.coordinates(0) + cam.coordinates(1)` gives a 2-dimensional array of coordinates.

        Args:
            dimension: Dimension for which to return the coordinates.
        """
        unit = u.dimensionless_unscaled if self.pixel_size is None else self.pixel_size[dimension]
        shape = np.ones_like(self.data_shape)
        shape[dimension] = self.data_shape[dimension]
        return np.arange(0.5, 0.5 + self.data_shape[dimension], 1.0).reshape(shape) * unit

    @final
    @property
    def extent(self) -> Quantity:
        """Physical size of the data array

        If a `pixel_size` is set, this function returns `data_shape * pixel_size`
        as an `astropy.units.Quantity`.
        If no `pixel_size` is set, this function uses  the
        `dimensionless_unscaled` unit.
        """
        unit = u.dimensionless_unscaled if self.pixel_size is None else self.pixel_size
        return np.array(self.data_shape) * unit


class Processor(Detector, ABC):
    """Base class for all Processors.

    Processors can be used to build data processing graphs, where each Processor takes input from one or
    more input Detectors and processes that data (e.g., cropping an image, averaging over an ROI, etc.).
    A processor, itself, is a Detector to allow chaining multiple processors together to combine functionality.

    To implement a processor, implement `_fetch`, and optionally override `data_shape`, `pixel_size`, and `__init__`.
    The `latency` and `duration` properties are computed from the latency and duration of the inputs and cannot be set.
    By default, the `pixel_size` and `data_shape` are the same as the `pixel_size` and `data_shape` of the first input.
    To override this behavior, override the `pixel_size` and `data_shape` properties.

    Args:
        multi_threaded: If True, `_fetch` is called from a worker thread. Otherwise, `_fetch` is called
            directly from `trigger`. If the device is not thread-safe, or threading provides no benefit,
            or for easy debugging, set this to False.
    """

    def __init__(self, *args, multi_threaded: bool):
        self._sources = args
        # data_shape, duration, latency and pixel_size all may change dynamically
        # when the settings of one of the source detectors is changed.
        # Therefore, we pass 'None' for all parameters, and override
        # data_shape, pixel_size, duration and latency in the properties.
        super().__init__(
            data_shape=None,
            pixel_size=None,
            duration=None,
            latency=None,
            multi_threaded=multi_threaded,
        )

    def trigger(self, *args, immediate=False, **kwargs):
        """Triggers all sources at the same time (regardless of latency), and schedules a call to `_fetch()`"""
        future_data = [
            (source.trigger(immediate=immediate) if source is not None else None) for source in self._sources
        ]
        return super().trigger(*future_data, *args, **kwargs)

    @final
    @property
    def latency(self) -> Quantity[u.ms]:
        """Returns the shortest latency for all detectors."""
        return min(
            (source.latency for source in self._sources if source is not None),
            default=0.0 * u.ms,
        )

    @final
    @property
    def duration(self) -> Quantity[u.ms]:
        """Returns the last end time minus the first start time for all detectors
        i.e., max (duration + latency) - min(latency).

        Note that `latency` is allowed to vary over time for devices that can only be triggered periodically,
        so this `duration` may also vary over time.
        """
        times = [(source.duration, source.latency) for source in self._sources if source is not None]
        if len(times) == 0:
            return 0.0 * u.ms
        return max([duration + latency for (duration, latency) in times]) - min(
            [latency for (duration, latency) in times]
        )

    @property
    def data_shape(self):
        """This default implementation returns the data shape of the first source."""
        return self._sources[0].data_shape

    @property
    def pixel_size(self) -> Optional[Quantity]:
        """This default implementation returns the pixel size of the first source."""
        return self._sources[0].pixel_size


class PhaseSLM(ABC):
    """Base class for phase-only Spatial Light Modulators (SLMs).

    This abstract base class defines the interface for phase-only spatial light modulators,
    which are devices that can modulate the phase of light. Implementations of this class
    should provide methods to set phase patterns and update the display.
    """

    __slots__ = ()

    @abstractmethod
    def update(self) -> None:
        """Sends the new phase pattern to be displayed on the SLM.

        Implementations should call _start() before triggering the SLM.

        Note:
            This function *does not* wait for the image to appear on the SLM.
            To wait for the image stabilization explicitly, use 'wait()'.
            However, this should rarely be needed since all Detectors
            already wait for the image to stabilize before starting a measurement.
        """

    @abstractmethod
    def set_phases(self, values: ArrayLike, update: bool = True) -> None:
        """Sets the phase pattern on the SLM.

        Args:
            values(ArrayLike): phase pattern, in radians.
                The pattern is automatically stretched to fill the full SLM.
            update: when True, calls `update` after setting the phase pattern.
                Set to `False` to suppress the call to `update`.
                This is useful in advanced scenarios where multiple parameters of the SLM need to be changed
                before updating the displayed image.
        """
        ...



import logging  # noqa

import astropy.units as u
import numpy as np
import pytest
import skimage

from openwfs.algorithms import StepwiseSequential
from openwfs.processors import SingleRoi
from openwfs.simulation import Microscope, Camera, StaticSource, SLM
from openwfs.utilities.patterns import tilt


def test_mock_camera_and_single_roi():
    """
    The MockCamera wraps a Detector producing 2-D data, so that the data can be read by MicroManager.
    The data is converted into a 16-bit int, and methods are added to set width, height, top and bottom.
    By default, the MockCamera scales the input image such that the maximum value is mapped to the unsigned integer
    0xFFFF = (2 ** 16) - 1.
    """
    img = np.zeros((1000, 1000), dtype=np.int16)
    img[200, 300] = 39.39  # some random float
    src = Camera(StaticSource(img, pixel_size=450 * u.nm), analog_max=None)
    roi_detector = SingleRoi(src, pos=(200, 300), radius=0)  # Only measure that specific point
    assert roi_detector.read() == int(2**16 - 1)  # it should cast the array into some int


@pytest.mark.parametrize("shape", [(1000, 1000), (999, 999)])
def test_microscope_without_magnification(shape):
    """
    Checks if the microscope can be constructed and read out without exceptions being thrown.
    Also checks if the microscope does not offset the image (for odd and even number of pixels)
    """
    # construct input image
    img = np.zeros(shape, dtype=np.int16)
    img[256, 256] = 100
    src = Camera(StaticSource(img, pixel_size=400 * u.nm), analog_max=0xFFFF)

    # construct microscope
    sim = Microscope(source=src, magnification=1, numerical_aperture=1, wavelength=800 * u.nm)
    cam = Camera(sim, analog_max=None)
    img = cam.read()
    assert img[256, 256] == 2**16 - 1


def test_microscope_and_aberration():
    """
    This test concerns the basic effect of casting an aberration or SLM pattern on the back pupil.
    """
    img = np.zeros((1000, 1000), dtype=np.int16)
    img[256, 256] = 100
    src = Camera(StaticSource(img, pixel_size=400 * u.nm), analog_max=None)

    slm = SLM(shape=(512, 512))

    aberrations = skimage.data.camera() * ((2 * np.pi) / 255.0)

    sim = Microscope(
        source=src,
        magnification=1,
        incident_field=slm.field,
        numerical_aperture=1,
        wavelength=800 * u.nm,
    )

    without_aberration = sim.read()[256, 256]
    slm.set_phases(aberrations)
    with_aberration = sim.read()[256, 256]
    assert with_aberration < without_aberration


def test_slm_and_aberration():
    """
    As mentioned in the previous test, casting a pattern on the pupil plane with an SLM and an aberration
    should produce the same effect. We will test that here by projecting two opposing patterns on the pupil plane.
    (Which should do nothing in the image plane)
    """
    img = np.zeros((1000, 1000), dtype=np.int16)
    img[256, 256] = 100
    src = Camera(StaticSource(img, pixel_size=400 * u.nm), analog_max=None)

    slm = SLM(shape=(512, 512))

    aberrations = skimage.data.camera() * ((2 * np.pi) / 255.0) * 0
    slm.set_phases(-aberrations)
    aberration = StaticSource(aberrations, pixel_size=1.0 / 512 * u.dimensionless_unscaled)

    sim1 = Microscope(
        source=src,
        incident_field=slm.field,
        numerical_aperture=1.0,
        aberrations=aberration,
        wavelength=800 * u.nm,
    )
    sim2 = Microscope(source=src, numerical_aperture=1.0, wavelength=800 * u.nm)

    # We correlate the two.
    # Any discrepancy between the two matrices should throw an error
    # try putting one of the wavelengths to 800

    a = sim1.read()
    b = sim2.read()
    norm_a = a / np.linalg.norm(a[:])
    norm_b = b / np.linalg.norm(b[:])

    assert abs(np.vdot(norm_a, norm_b)) >= 1


def test_slm_tilt():
    """
    Display a tilt on the SLM should result in an image plane shift. If the magnification is 1, this should
    correspond to a tilt of 1 pixel for a 2 pi phase shift.
    """
    img = np.zeros((1000, 1000), dtype=np.int16)
    signal_location = (256, 250)
    img[signal_location] = 100
    pixel_size = 400 * u.nm
    wavelength = 750 * u.nm
    src = Camera(StaticSource(img, pixel_size=pixel_size), analog_max=None)

    slm = SLM(shape=(1000, 1000))

    na = 1.0
    sim = Microscope(
        source=src,
        incident_field=slm.field,
        magnification=1,
        numerical_aperture=na,
        wavelength=wavelength,
    )

    # introduce a tilted pupil plane
    # the input parameter to `tilt` corresponds to a shift 2.0/π the Abbe diffraction limit.
    shift = np.array((-24, 40))
    step = wavelength / (np.pi * na)
    slm.set_phases(tilt(1000, -shift * pixel_size / step))

    new_location = signal_location + shift

    cam = Camera(sim, analog_max=None)
    img = cam.read(immediate=True)
    max_pos = np.unravel_index(np.argmax(img), img.shape)
    assert np.all(max_pos == new_location)


def test_microscope_wavefront_shaping(caplog):
    """
    Reproduces a bug that occurs due to the location of the measurements.wait() command.
    """
    # caplog.set_level(logging.DEBUG)
    aberrations = skimage.data.camera() * ((2 * np.pi) / 255.0) + np.pi

    aberration = StaticSource(aberrations, pixel_size=1.0 / 512 * u.dimensionless_unscaled)  # note: incorrect scaling!

    img = np.zeros((1000, 1000), dtype=np.int16)
    img[256, 256] = 100

    img = np.zeros((1000, 1000), dtype=np.int16)
    signal_location = (250, 200)

    img[signal_location] = 100
    src = StaticSource(img, pixel_size=400 * u.nm)

    slm = SLM(shape=(1000, 1000))

    sim = Microscope(
        source=src,
        incident_field=slm.field,
        numerical_aperture=1,
        aberrations=aberration,
        wavelength=800 * u.nm,
    )

    cam = Camera(sim, analog_max=100)
    roi_detector = SingleRoi(cam, pos=signal_location, radius=0)  # Only measure that specific point

    alg = StepwiseSequential(feedback=roi_detector, slm=slm, phase_steps=3, n_x=3, n_y=3)
    t = alg.execute().t

    # test if the modes differ. The error causes them not to differ
    assert np.std(t[:][1:]) > 0


def phase_response_test_function(phi, b, c, gamma):
    """A synthetic phase response function: 2π*(b + c*(phi/2π)^gamma)"""
    return np.clip(2 * np.pi * (b + c * (phi / (2 * np.pi)) ** gamma), 0, None)


def inverse_phase_response_test_function(f, b, c, gamma):
    """Inverse of the synthetic phase response function: 2π*(b + c*(phi/2π)^gamma)"""
    return 2 * np.pi * ((f / (2 * np.pi) - b) / c) ** (1 / gamma)


def lookup_table_test_function(f, b, c, gamma):
    """
    Compute the lookup indices (i.e. a lookup table)
    for countering the synthetic phase response test function: 2π*(b + c*(phi/2π)^gamma).
    """
    phase = inverse_phase_response_test_function(f, b, c, gamma)
    return (np.mod(phase, 2 * np.pi) * 256 / (2 * np.pi) + 0.5).astype(np.uint8)


def test_mock_slm_lut_and_phase_response():
    """
    Test the lookup table and phase response of the MockSLM.
    """
    # === Test default lookup table and phase response ===
    # Includes edge cases like rounding/wrapping: -0.501 -> 255, -0.499 -> 0
    input_phases_a = (
        np.asarray(
            (
                -1,
                -0.501,
                -0.499,
                0,
                1,
                64,
                128,
                192,
                255,
                255.499,
                255.501,
                256,
                257,
                511,
                512,
            )
        )
        * 2
        * np.pi
        / 256
    )
    expected_output_phases_a = (
        np.asarray((255, 255, 0, 0, 1, 64, 128, 192, 255, 255, 0, 0, 1, 255, 0)) * 2 * np.pi / 256
    )
    slm1 = SLM(shape=(3, input_phases_a.shape[0]))
    slm1.set_phases(input_phases_a)
    assert np.all(np.abs(slm1.phases.read() - expected_output_phases_a) < 1e6)

    # === Test setting custom phase response of SLM ===
    # Construct custom phase response
    b = -0.02
    c = 1.2
    gamma = 1.6
    linear_phase = np.arange(0, 2 * np.pi, 2 * np.pi / 256)
    phase_response = phase_response_test_function(linear_phase, b, c, gamma)

    # Apply and test custom synthetic phase response
    slm2 = SLM(shape=(3, 256))
    slm2.phase_response = phase_response
    slm2.set_phases(linear_phase)
    assert np.all(np.abs(slm2.phases.read() - phase_response) < 1e-6)

    # === Test custom lookup table ===
    lookup_table = lookup_table_test_function(linear_phase, b, c, gamma)
    slm3 = SLM(shape=(3, 256))
    slm3.lookup_table = lookup_table
    slm3.set_phases(linear_phase)
    assert np.all(
        np.abs(slm3.phases.read() - inverse_phase_response_test_function(linear_phase, b, c, gamma))
        < (1.1 * np.pi / 256)
    )

    # === Test custom lookup table that counters custom synthetic phase response ===
    linear_phase_highres = np.arange(0, 2 * np.pi * 255.49 / 256, 0.25 * 2 * np.pi / 256)
    slm4 = SLM(shape=(3, linear_phase_highres.shape[0]))
    slm4.phase_response = phase_response
    slm4.lookup_table = lookup_table
    slm4.set_phases(linear_phase_highres)
    assert np.all(np.abs(slm4.phases.read()[0] - linear_phase_highres) < (3 * np.pi / 256))



import astropy.units as u
import numpy as np
import pytest

from . import complex_random
from .test_simulation import phase_response_test_function, lookup_table_test_function
from openwfs.algorithms import StepwiseSequential
from openwfs.algorithms.troubleshoot import (
    cnr,
    signal_std,
    find_pixel_shift,
    field_correlation,
    frame_correlation,
    pearson_correlation,
    measure_modulated_light,
    measure_modulated_light_dual_phase_stepping,
)
from openwfs.algorithms.utilities import analyze_phase_stepping
from openwfs.processors import SingleRoi
from openwfs.simulation import Camera
from openwfs.simulation import SimulatedWFS, StaticSource, SLM, Microscope


@pytest.mark.parametrize("phase_steps", [5, 6, 10, 20])
@pytest.mark.parametrize("noise", [0.0, 0.5, 1.0, 2.0, 4.0])
def test_analyze_phase_stepping(phase_steps, noise):
    """Test the analyze_phase_stepping function"""
    # TODO: find out why there is a (small) systematic error when the noise is high
    # Construct a perfect phase stepping signal
    np.random.seed(123)
    t = complex_random((10000, 1))
    ref = np.exp(np.arange(phase_steps) * 2j * np.pi / phase_steps)
    signal = np.abs(ref + t) ** 2
    signal += np.random.normal(scale=noise, size=signal.shape)
    result = analyze_phase_stepping(signal, axis=1)
    # the signal energy for a signal 2·cos(φ) is 2 P  (with P the number of phase steps), distributed over 2 bins,
    # giving P per bin.
    # the noise energy is σ²·P, distributed over P bins, giving σ² per bin.
    tolerance = 4 / np.sqrt(t.size)
    theoretical_fidelity = phase_steps / (phase_steps + noise**2)
    print(f"noise fidelity. theoretical: {theoretical_fidelity}  measured: {result.fidelity_noise}")
    assert np.isclose(result.fidelity_noise, theoretical_fidelity, rtol=tolerance)


def test_signal_std():
    """
    Test signal std, corrected for (uncorrelated) noise in the signal.
    """
    a = np.random.rand(400, 400)
    b = np.random.rand(400, 400)
    assert signal_std(a, a) < 1e-6  # Test noise only
    assert np.abs(signal_std(a + b, b) - a.std()) < 0.005  # Test signal+uncorrelated noise
    assert np.abs(signal_std(a + a, a) - np.sqrt(3) * a.std()) < 0.005  # Test signal+correlated noise


def test_cnr():
    """
    Test Contrast to Noise Ratio, corrected for (uncorrelated) noise in the signal.
    """
    a = np.random.randn(800, 800)
    b = np.random.randn(800, 800)
    cnr_gt = 3.0  # Ground Truth
    assert cnr(a, a) < 1e-6  # Test noise only
    assert np.abs(cnr(cnr_gt * a + b, b) - cnr_gt) < 0.01  # Test signal+uncorrelated noise
    assert np.abs(cnr(cnr_gt * a + a, a) - np.sqrt((cnr_gt + 1) ** 2 - 1)) < 0.01  # Test signal+correlated noise


def test_find_pixel_shift():
    """
    Test finding pixel shifts between two images.
    """
    # Define pixel shifts to test
    ab_shift = (-3, -5)
    cd_shift = (2, -4)

    # Define random image 2D arrays
    a = np.random.rand(20, 19)
    b = np.roll(a, shift=ab_shift, axis=(0, 1))
    c = np.random.rand(17, 18)
    d = np.roll(c, shift=cd_shift, axis=(0, 1))

    # Find pixel shifts
    aa_shift_found = find_pixel_shift(a, a)
    ab_shift_found = find_pixel_shift(a, b)
    cc_shift_found = find_pixel_shift(c, c)
    cd_shift_found = find_pixel_shift(c, d)

    # Assert shifts
    assert aa_shift_found == (0, 0)
    assert ab_shift_found == ab_shift
    assert cc_shift_found == (0, 0)
    assert cd_shift_found == cd_shift


def test_field_correlation():
    """
    Test the field correlation, i.e. g_1 normalized first order correlation function.
    """
    a = np.zeros(shape=(2, 3))
    a[1, 2] = 2.0
    a[0, 0] = 0.0

    b = np.ones(shape=(2, 3))
    b[1, 2] = 0.0
    b[0, 0] = 0.0

    c = np.ones(shape=(2, 3), dtype=np.cdouble)
    c[1, 0] = 1 + 1j
    c[0, 1] = 2 - 3j

    assert field_correlation(a, a) == 1.0  # Self-correlation
    assert field_correlation(2 * a, a) == 1.0  # Invariant under scalar-multiplication
    assert field_correlation(a, b) == 0.0  # Orthogonal arrays
    assert np.abs(field_correlation(a + b, b) - np.sqrt(0.5)) < 1e-10  # Self+orthogonal array
    assert np.abs(field_correlation(b, c) - np.conj(field_correlation(c, b))) < 1e-10  # Arguments swapped


def test_frame_correlation():
    """
    Test the frame correlation, i.e. g_2 normalized second order correlation function.
    Test the following:
        g_2 correlation with self == 1/3 for distribution from `random.rand`
        g_2 correlation with other == 0
    """
    a = np.random.rand(1000000)
    b = np.random.rand(1000000)

    assert np.abs(frame_correlation(a, a) - 1 / 3) < 2e-3
    assert np.abs(frame_correlation(a, b)) < 2e-3


def test_pearson_correlation():
    """
    Test the Pearson correlation.
    """
    # Perfect correlation
    a = np.asarray((1, 2, 3))
    b = np.asarray((2, 4, 6))
    corr_ab = pearson_correlation(a, b)
    corr_minus_ab = pearson_correlation(-a, b)
    assert np.isclose(corr_ab, 1, atol=1e-6)
    assert np.isclose(corr_minus_ab, -1, atol=1e-6)

    # No correlation
    c = np.asarray((1, 0, -1))
    d = np.asarray((0, 2, 0))
    corr_cd_compute = pearson_correlation(c, d)
    assert np.isclose(corr_cd_compute, 0, atol=1e-6)

    # Some correlation
    e = np.asarray((4, -4, 2, -2))
    f = np.asarray((2, -2, -1, 1))
    corr_ef = pearson_correlation(e, f)
    assert np.isclose(corr_ef, 0.6, atol=1e-6)


def test_pearson_correlation_noise_compensated():
    """
    Test Pearson correlation, compensated for noise.

    For Spearman's attenuation factor, see also:
     - Saccenti et al. 2020, https://www.nature.com/articles/s41598-019-57247-4
     - https://en.wikipedia.org/wiki/Regression_dilution#Correlation_correction
    """
    N = 1000000
    a = np.random.rand(N)
    b = np.random.rand(N)
    noise1 = np.random.rand(N)
    noise2 = np.random.rand(N)

    # Generate fake signals
    A1 = 3 * a
    A2 = 4 * a
    B2 = 5 * b
    A_noisy1 = A1 + noise1
    A_noisy2 = A2 + noise2
    B_noisy2 = B2 + noise2

    corr_AA = pearson_correlation(A_noisy1, A_noisy2, noise_var=noise1.var())
    corr_AB = pearson_correlation(A_noisy1, B_noisy2, noise_var=noise1.var())
    corr_AA_with_noise = pearson_correlation(A_noisy1, A_noisy2)

    assert np.isclose(noise1.var(), noise2.var(), atol=2e-3)
    assert np.isclose(corr_AA, 1, atol=2e-3)
    assert np.isclose(corr_AB, 0, atol=2e-3)
    A_spearman = 1 / np.sqrt((1 + noise1.var() / A1.var()) * (1 + noise2.var() / A2.var()))
    assert np.isclose(corr_AA_with_noise, A_spearman, atol=2e-3)


@pytest.mark.parametrize(
    "n_y, n_x, phase_steps, b, c, gamma",
    [(11, 9, 8, -0.05, 1.5, 0.8), (4, 4, 10, -0.05, 1.5, 0.8)],
)
def test_fidelity_phase_calibration_ssa_noise_free(n_y, n_x, phase_steps, b, c, gamma):
    """
    Test computing phase calibration fidelity factor, with the SSA algorithm. Noise-free scenarios.
    """
    # Perfect SLM, noise-free
    aberrations = np.random.uniform(0.0, 2 * np.pi, (n_y, n_x))
    sim = SimulatedWFS(aberrations=aberrations)
    alg = StepwiseSequential(feedback=sim, slm=sim.slm, n_x=n_x, n_y=n_y, phase_steps=phase_steps)
    result = alg.execute()
    assert result.fidelity_calibration > 0.99

    # SLM with incorrect phase response, noise-free
    linear_phase = np.arange(0, 2 * np.pi, 2 * np.pi / 256)
    sim.slm.phase_response = phase_response_test_function(linear_phase, b, c, gamma)
    result = alg.execute()
    assert result.fidelity_calibration < 0.9

    # SLM calibrated with phase response corrected by LUT, noise-free
    sim.slm.lookup_table = lookup_table_test_function(linear_phase, b, c, gamma)
    result = alg.execute()
    assert result.fidelity_calibration > 0.99


@pytest.mark.parametrize("n_y, n_x, phase_steps, gaussian_noise_std", [(4, 4, 10, 0.2), (6, 6, 12, 1.0)])
def test_fidelity_phase_calibration_ssa_with_noise(n_y, n_x, phase_steps, gaussian_noise_std):
    """
    Test estimation of phase calibration fidelity factor, with the SSA algorithm. With noise.
    """
    # === Define mock hardware, perfect SLM ===
    # Aberration and image source
    numerical_aperture = 1.0
    aberration_phase = np.random.uniform(0.0, 2 * np.pi, (n_y, n_x))
    aberration = StaticSource(aberration_phase, extent=2 * numerical_aperture)
    img = np.zeros((64, 64), dtype=np.int16)
    img[32, 32] = 250
    src = StaticSource(img, pixel_size=500 * u.nm)

    # SLM, simulation, camera, ROI detector
    slm = SLM(shape=(80, 80))
    sim = Microscope(
        source=src,
        incident_field=slm.field,
        magnification=1,
        numerical_aperture=numerical_aperture,
        aberrations=aberration,
        wavelength=800 * u.nm,
    )
    cam = Camera(sim, analog_max=1e4, gaussian_noise_std=gaussian_noise_std)
    roi_detector = SingleRoi(cam, radius=0)  # Only measure that specific point

    # Define and run WFS algorithm
    alg = StepwiseSequential(feedback=roi_detector, slm=slm, n_x=n_x, n_y=n_y, phase_steps=phase_steps)
    result_good = alg.execute()
    assert result_good.fidelity_calibration > 0.9

    # SLM with incorrect phase response
    linear_phase = np.arange(0, 2 * np.pi, 2 * np.pi / 256)
    slm.phase_response = phase_response_test_function(linear_phase, b=0.05, c=0.6, gamma=1.5)
    result_good = alg.execute()
    assert result_good.fidelity_calibration < 0.9


@pytest.mark.parametrize("num_blocks, phase_steps, expected_fid, atol", [(10, 8, 1, 1e-6)])
def test_measure_modulated_light_dual_phase_stepping_noise_free(num_blocks, phase_steps, expected_fid, atol):
    """Test fidelity estimation due to amount of modulated light. Noise-free."""
    # Perfect SLM, noise-free
    aberrations = np.random.uniform(0.0, 2 * np.pi, (20, 20))
    sim = SimulatedWFS(aberrations=aberrations)

    # Measure the amount of modulated light (no non-modulated light present)
    fidelity_modulated = measure_modulated_light_dual_phase_stepping(
        slm=sim.slm, feedback=sim, phase_steps=phase_steps, num_blocks=num_blocks
    )
    assert np.isclose(fidelity_modulated, expected_fid, atol=atol)


@pytest.mark.parametrize(
    "num_blocks, phase_steps, gaussian_noise_std, atol",
    [(10, 6, 0.0, 1e-6), (6, 8, 2.0, 1e-3)],
)
def test_measure_modulated_light_dual_phase_stepping_with_noise(num_blocks, phase_steps, gaussian_noise_std, atol):
    """Test fidelity estimation due to amount of modulated light. Can test with noise."""
    # === Define mock hardware, perfect SLM ===
    # Aberration and image source
    img = np.zeros((64, 64), dtype=np.int16)
    img[32, 32] = 100
    src = StaticSource(img, pixel_size=200 * u.nm)

    # SLM, simulation, camera, ROI detector
    slm = SLM(shape=(100, 100))
    sim = Microscope(
        source=src,
        incident_field=slm.field,
        magnification=1,
        numerical_aperture=1.0,
        wavelength=800 * u.nm,
    )
    cam = Camera(sim, analog_max=1e4, gaussian_noise_std=gaussian_noise_std)
    roi_detector = SingleRoi(cam, radius=0)  # Only measure that specific point

    # Measure the amount of modulated light (no non-modulated light present)
    fidelity_modulated = measure_modulated_light_dual_phase_stepping(
        slm=slm, feedback=roi_detector, phase_steps=phase_steps, num_blocks=num_blocks
    )
    assert np.isclose(fidelity_modulated, 1, atol=atol)


@pytest.mark.parametrize(
    "phase_steps, modulated_field_amplitude, non_modulated_field",
    [(6, 1.0, 0.0), (8, 0.5, 0.5), (8, 1.0, 0.25)],
)
def test_measure_modulated_light_noise_free(phase_steps, modulated_field_amplitude, non_modulated_field):
    """Test fidelity estimation due to amount of modulated light. Noise-free."""
    # Perfect SLM, noise-free
    aberrations = np.random.uniform(0.0, 2 * np.pi, (20, 20))
    slm = SLM(
        aberrations.shape,
        field_amplitude=modulated_field_amplitude,
        non_modulated_field_fraction=non_modulated_field,
    )
    sim = SimulatedWFS(aberrations=aberrations, slm=slm)

    # Measure the amount of modulated light (no non-modulated light present)
    fidelity_modulated = measure_modulated_light(slm=sim.slm, feedback=sim, phase_steps=phase_steps)
    expected_fid = 1.0 / (1.0 + non_modulated_field**2)
    assert np.isclose(fidelity_modulated, expected_fid, rtol=0.1)


@pytest.mark.parametrize(
    "phase_steps, gaussian_noise_std, modulated_field_amplitude, non_modulated_field",
    [(8, 0.0, 0.5, 0.4), (6, 0.0, 1.0, 0.0), (12, 2.0, 1.0, 0.25)],
)
def test_measure_modulated_light_dual_phase_stepping_with_noise(
    phase_steps, gaussian_noise_std, modulated_field_amplitude, non_modulated_field
):
    """Test fidelity estimation due to amount of modulated light. Can test with noise."""
    # === Define mock hardware, perfect SLM ===
    # Aberration and image source
    img = np.zeros((64, 64), dtype=np.int16)
    img[32, 32] = 100
    src = StaticSource(img, pixel_size=200 * u.nm)

    # SLM, simulation, camera, ROI detector
    slm = SLM(
        shape=(100, 100),
        field_amplitude=modulated_field_amplitude,
        non_modulated_field_fraction=non_modulated_field,
    )
    sim = Microscope(source=src, incident_field=slm.field, wavelength=800 * u.nm)
    cam = Camera(sim, analog_max=1e3, gaussian_noise_std=gaussian_noise_std)
    roi_detector = SingleRoi(cam, radius=0)  # Only measure that specific point

    # Measure the amount of modulated light (no non-modulated light present)
    expected_fid = 1.0 / (1.0 + non_modulated_field**2)
    fidelity_modulated = measure_modulated_light(slm=slm, feedback=roi_detector, phase_steps=phase_steps)
    assert np.isclose(fidelity_modulated, expected_fid, rtol=0.1)



import astropy.units as u
import harvesters
import pytest

from openwfs.devices import Camera, is_loaded
from openwfs.processors import HDRCamera

if not is_loaded(harvesters):
    pytest.skip(harvesters.message, allow_module_level=True)


cti_path = R"C:\Program Files\Basler\pylon 7\Runtime\x64\ProducerU3V.cti"


@pytest.fixture
def camera():
    """
    Fixture that returns a Camera object.
    If no camera is found, the test is skipped.
    """
    cameras = Camera.enumerate_cameras(cti_path)
    if len(cameras) == 0:
        pytest.skip("No camera found", allow_module_level=True)
    return Camera(cti_path)


def test_grab(camera):
    frame = camera.read()
    assert frame.shape == camera.data_shape
    assert (camera.height, camera.width) == camera.data_shape
    assert (camera._nodes.Height.value, camera._nodes.Width.value) == camera.data_shape


def test_hdr(camera):
    """Test the HDR camera functionality.
    Note: this just tests if the code can be run, visual inspection is needed to see if the result is correct."""
    camera.exposure = 0.1 * u.s
    hdr_camera = HDRCamera(camera, exposure_factors=(1.0, 0.1, 0.01), background=0, saturation_threshold=255)
    hdr_camera.read()

    # import matplotlib.pyplot as plt
    # first = True
    # for i in range(0):
    #     hdr_frame = hdr_camera.read()
    #     frame = camera.read()
    #     plt.subplot(121)
    #     plt.title("Original")
    #     plt.imshow(frame)
    #     if first:
    #         plt.colorbar()
    #
    #     plt.subplot(122)
    #     plt.title("HDR")
    #     plt.imshow(hdr_frame)
    #     plt.show(block=False)
    #     if first:
    #         plt.colorbar()
    #         first = False
    #     plt.pause(0.1)


@pytest.mark.parametrize("binning", [1, 2, 4])
@pytest.mark.parametrize("top, left", [(0, 0), (21, 17)])
def test_roi(camera, binning, top, left):
    original_shape = camera.data_shape

    # check if binning returns the correct shape
    # take care that the size will be a multiple of the increment,
    # and that setting the binning will round this number down
    camera.binning = binning
    expected_width = (original_shape[1] // binning) // camera._nodes.Width.inc * camera._nodes.Width.inc
    expected_height = (original_shape[0] // binning) // camera._nodes.Height.inc * camera._nodes.Height.inc
    assert camera.data_shape == (expected_height, expected_width)

    # check if setting the ROI works
    new_height = camera.height // 2
    new_width = camera.width // 2
    camera.height = new_height
    camera.width = new_width
    camera.top = top
    camera.left = left

    frame = camera.read()
    assert frame.shape == camera.data_shape
    assert (camera.height, camera.width) == camera.data_shape
    assert (camera._nodes.Height.value, camera._nodes.Width.value) == camera.data_shape

    # the camera is allowed to round up to multiples of 16
    # check if camera.width - new_width is in the range 0 to 15
    assert 0 <= camera.width - new_width < camera._nodes.Width.inc
    assert 0 <= camera.height - new_height < camera._nodes.Height.inc



import time

import astropy.units as u
import cv2
import glfw
import numpy as np  # for debugging
import pytest

from openwfs.devices import is_loaded
from openwfs.devices.slm import SLM, Patch, geometry
from openwfs.utilities import Transform

if not is_loaded(glfw):
    pytest.skip(glfw.message, allow_module_level=True)

glfw.init()
if not glfw.get_monitors():
    pytest.skip("No monitors found", allow_module_level=True)


# just some values for testing different gray value levels
GVAL1 = 173
GVAL2 = 15
GVAL3 = 23
VAL1 = 2 * np.pi / 256 * GVAL1
VAL2 = 2 * np.pi / 256 * GVAL2


@pytest.fixture
def slm() -> SLM:
    slm = SLM(monitor_id=0, shape=(100, 200), pos=(20, 10), coordinate_system="full")
    return slm


def test_create_windowed(slm):
    """Tests creating a windowed SLM. We should be able to make multiple windowed SLMs, but only one full screen SLM."""

    # check if size and position match the specification
    assert slm.shape == (100, 200)
    assert slm.position == (20, 10)
    assert slm.transform == Transform.identity()
    assert slm.coordinate_system == "full"

    # check if frame buffer has correct size
    fb_texture = slm._frame_buffer._textures[Patch._PHASES_TEXTURE]
    assert fb_texture.get_data().shape == slm.shape

    # check if the refresh rate got set
    assert int(slm.refresh_rate.to_value(u.Hz)) != glfw.DONT_CARE

    # move window (we cannot really check this, but we just check if no error occurs)
    slm.position = (2, 1)
    assert slm.position == (2, 1)

    # we should be able to construct a second windowed SLM
    _slm2 = SLM(0)

    # but not a full screen SLM on this window
    with pytest.raises(RuntimeError):
        SLM(1)

    # and we also expect an error when the window does not exist
    with pytest.raises(IndexError):
        SLM(999)

    # we also cannot make the slm full screen, because slm2 is still active
    with pytest.raises(RuntimeError):
        slm.monitor_id = 1

    # but if we delete slm2 first, we can make the slm full screen
    _slm2 = None
    slm.monitor_id = 1
    full_screen_shape = slm.shape
    assert full_screen_shape[0] > 200  # assuming we don't have a tiny monitor
    assert full_screen_shape[1] > 200  # assuming we don't have a tiny monitor
    assert slm.refresh_rate.to_value(u.Hz) != glfw.DONT_CARE

    # we cannot make a windowed slm now because the full screen is already taken
    with pytest.raises(RuntimeError):
        slm3 = SLM(0)  # noqa: not used, just to check if an error occurs

    # make the slm windowed again
    # at the moment, the size of the window does not need to be conserved
    slm.monitor_id = 0
    assert slm.shape[0] < full_screen_shape[0]
    assert slm.shape[1] < full_screen_shape[1]

    # we can now make a second windowed slm
    slm4 = SLM(0)  # noqa: not used, just to check if an error occurs


# noinspection DuplicatedCode
def test_transform(slm):
    """Tests the transform property of the SLM class."""

    # construct a test pattern:
    # 1 | 1 | 2 | 2
    # --+---+---+--
    # 0 | 0 | 3 | 3
    # --+---+---+--
    # 0 | 0 | 3 | 3
    # --+---+---+--
    # 0 | 0 | 3 | 3
    pattern = np.array((1, 1, 2, 2, 0, 0, 3, 3, 0, 0, 3, 3, 0, 0, 3, 3)).reshape((4, 4))
    assert pattern[1, 3] == 3

    # display the pattern on the SLM, using the 'full' transform
    slm.set_phases(2.0 * np.pi * pattern / 4.0)

    # check if the pattern is displayed correctly
    pixels = slm.pixels.read() / 64
    phases = slm.phases.read() * 4 / (2 * np.pi)
    assert np.allclose(pixels, phases)
    assert np.allclose(pixels[:25, :100], 1)
    assert np.allclose(pixels[25:, :100], 0)
    assert np.allclose(pixels[:25, 100:], 2)
    assert np.allclose(pixels[25:, 100:], 3)

    # now change the transform to 'short' to fit the pattern to a centered square, with the height of the
    # SLM.
    # Then check if the pattern is displayed correctly
    slm.coordinate_system = "short"  # does not trigger an update
    assert np.all(slm.pixels.read() / 64 == pixels)
    slm.update()

    pixels = slm.pixels.read() / 64
    phases = slm.phases.read() * 4 / (2 * np.pi)
    assert np.allclose(pixels, phases)
    assert np.allclose(pixels[:, :50], 0)
    assert np.allclose(pixels[:, 150:], 0)
    assert np.allclose(pixels[:25, 50:100], 1)
    assert np.allclose(pixels[25:, 50:100], 0)
    assert np.allclose(pixels[:25, 100:150], 2)
    assert np.allclose(pixels[25:, 100:150], 3)

    # now change the transform to 'long' to fit the pattern to a centered square, with the width of the
    # SLM, causing part of the texture to be mapped outside the window.
    slm.coordinate_system = "long"  # does not trigger an update
    assert np.all(slm.pixels.read() / 64 == pixels)
    slm.update()

    pixels = slm.pixels.read() / 64
    phases = slm.phases.read() * 4 / (2 * np.pi)
    assert np.allclose(pixels, phases)
    assert np.allclose(pixels[:, :100], 0)
    assert np.allclose(pixels[:, 100:], 3)

    # test zooming the pattern
    slm.coordinate_system = "short"
    slm.transform = Transform.zoom(0.8)
    slm.update()

    pixels = slm.pixels.read() / 64
    phases = slm.phases.read() * 4 / (2 * np.pi)
    assert np.allclose(pixels, phases)
    sub = pixels[10:90, 60:140].copy()
    pixels[10:90, 60:140] = 0
    assert np.allclose(pixels, 0)
    assert np.allclose(sub[:20, :40], 1)
    assert np.allclose(sub[20:, :40], 0)
    assert np.allclose(sub[:20, 40:], 2)
    assert np.allclose(sub[20:, 40:], 3)


@pytest.mark.skip(
    reason="This test is skipped by default because it causes the screen to flicker, which may "
    "affect people with epilepsy."
)
def test_refresh_rate():
    slm = SLM(1, latency=0, duration=0)
    refresh_rate = slm.refresh_rate
    slm.update()
    frame_count = 100

    # warm up
    for i in range(frame_count):
        slm.set_phases(0.0 if (i % 2) else np.pi)

    start = time.time_ns() * u.ns
    for i in range(frame_count):
        slm.set_phases(0.0 if (i % 2) else np.pi)
    stop = time.time_ns() * u.ns
    del slm
    actual_refresh_rate = frame_count / (stop - start)
    assert np.allclose(refresh_rate.to_value(u.Hz), actual_refresh_rate.to_value(u.Hz), rtol=1e-2)


def test_get_pixels():
    width = 73
    height = 99
    slm = SLM(SLM.WINDOWED, shape=(height, width))
    slm.coordinate_system = "full"  # fill full screen exactly (anisotropic coordinates
    pattern = np.random.uniform(size=(height, width)) * 2 * np.pi
    slm.set_phases(pattern)
    read_back = slm.pixels.read()
    diff = pattern - read_back / 256 * 2 * np.pi
    diff[diff > np.pi] -= 2 * np.pi
    assert np.allclose(diff, 0, atol=2 * np.pi / 256)

    read_back = slm.phases.read()
    diff = pattern - read_back
    diff[diff > np.pi] -= 2 * np.pi
    assert np.allclose(diff, 0, atol=1e-6)


def test_lookup_table(slm):
    # verify that we cannot set properties that do not exist on the object
    with pytest.raises(AttributeError):
        slm.lut = np.arange(256)  # noqa - the field 'lut' does not exist

    # resize the SLM and put a linear ramp on it
    slm.shape = (1, 256)
    slm.set_phases(np.arange(256).reshape(1, 256) * 2 * np.pi / 256)

    # read back the pixels and verify conversion to gray values
    pixels = slm.pixels.read()
    assert np.allclose(pixels, np.arange(256))

    # set a lookup table that is a random permutation of the gray values
    lut = np.arange(256)
    np.random.shuffle(lut)
    slm.lookup_table = lut

    # nothing changes until we call update
    assert np.all(pixels == slm.pixels.read())

    slm.update()
    pixels = slm.pixels.read()
    assert np.allclose(pixels, lut)

    # test phase wrapping
    slm.set_phases(np.arange(256).reshape(1, 256) * 2 * np.pi / 256 + 2 * np.pi)
    pixels_wrapped = slm.pixels.read()
    assert np.allclose(pixels, pixels_wrapped)


def test_multi_patch(slm):
    slm.shape = (2, 2)
    slm.patches.append(Patch(slm))
    pattern1 = np.array(((5, 7), (111, 13)))
    pattern2 = np.array(((29, 31), (137, 300)))
    slm.patches[0].set_phases(pattern1 * 2 * np.pi / 256)
    slm.patches[1].set_phases(pattern2 * 2 * np.pi / 256)

    # with additive_blend, the patterns should be added and wrapped to 2 pi (256)
    slm.patches[1].additive_blend = True
    slm.update()
    assert np.allclose(slm.pixels.read(), (pattern1 + pattern2) % 256)

    # without additive_blend, the second patch overwrites the pixels of the first one
    slm.patches[1].additive_blend = False
    slm.update()
    assert np.allclose(slm.pixels.read(), pattern2 % 256)

    # disable the second patch and check if the first one is still there
    slm.patches[1].enabled = False
    slm.update()
    assert np.allclose(slm.pixels.read(), pattern1 % 256)


def test_circular_geometry(slm):
    slm.shape = (200, 200)
    slm.set_phases(np.linspace(0, 2 * np.pi, 70, endpoint=False).reshape(1, -1))
    radii = (0.0, 0.25, 0.50, 1.0)
    segments = (10, 20, 40)
    slm.patches[0].geometry = geometry.circular(radii=radii, segments_per_ring=segments)
    slm.update()

    # read back the pixels and verify conversion to gray values
    pixels = np.rint(slm.pixels.read() / 256 * 70)
    polar_pixels = cv2.warpPolar(pixels, (100, 40), (99.5, 99.5), 100, cv2.WARP_POLAR_LINEAR)

    assert np.allclose(
        polar_pixels[:, 3:24],
        np.repeat(np.flip(np.arange(0, 10)), 4).reshape((-1, 1)),
        atol=1,
    )
    assert np.allclose(
        polar_pixels[:, 27:47],
        np.repeat(np.flip(np.arange(10, 30)), 2).reshape((-1, 1)),
        atol=1,
    )
    assert np.allclose(
        polar_pixels[:, 53:97],
        np.repeat(np.flip(np.arange(30, 70)), 1).reshape((-1, 1)),
        atol=1,
    )



default_vertex_shader = """
        #version 440 core
        layout(location = 0) in vec2 slm_coordinate;
        layout(location = 1) in vec2 texture_coordinate;
        layout(binding = 1, std140) uniform globals {
            mat3x2 transform;
        };
        out vec2 texCoord;

        void main() {
            vec2 pos = transform * vec3(slm_coordinate, 1.0);
            gl_Position = vec4(pos.x, pos.y, 0.0, 1.0);
            texCoord = texture_coordinate;
        }
        """

default_fragment_shader = """
        #version 440 core
        in vec2 texCoord;
        out vec4 colorOut;
        layout(binding = 0) uniform sampler2D texSampler;
        
        void main() {
            float val = texture(texSampler, texCoord).r;
            colorOut = vec4(val, val, val, 1.0);
        }
    """

# This shader is used to apply a software lookup table to phase image.
# The phase in radians is converted to a coordinate in the lookup table
# using the equation: texCoord = phase / (2 * pi) + 0.5 / 256
# where the offset is used for proper rounding, so that with the standard lookup table
# the range -δ to δ maps to a gray value 0 instead of
# negative values mapping to 255 and positive values mapping to 0.
# Since the lookup table texture is configured to use GL_WRAP,
# only the fractional part of texCoord is used (i.e., texCoord - floor(texCoord)).
#
post_process_fragment_shader = """
        #version 440 core
        in vec2 texCoord;
        out vec4 colorOut;
        layout(binding = 0) uniform sampler2D texSampler;
        layout(binding = 1) uniform sampler1D LUT;
        const float scale = 0.15915494309189535f; // corresponds to 1/(2 pi).
        const float offset = 0.001953125f; // corresponds to 0.5/256.

        void main() {
            float raw = texture(texSampler, texCoord).r * scale + offset;
            float val = texture(LUT, raw).r;
            colorOut = vec4(val, val, val, 1.0);
        }
    """

post_process_vertex_shader = """
        #version 440 core
        layout(location = 0) in vec2 slm_coordinate;
        layout(location = 1) in vec2 texture_coordinate;
        out vec2 texCoord;

        void main() {
            gl_Position = vec4(slm_coordinate, 0.0, 1.0);
            texCoord = texture_coordinate;
        }
    """



import threading
import weakref

import glfw

SLM = "slm.SLM"


class Context:
    """Holds a weak reference to an SLM window

    The context is used to activate the OpenGL context of the window.
    It can be used as `with context:` to activate the OpenGL context of the SLM window.
    Just to be sure, it also globally locks all access to OpenGL, so that only
     one thread can use OpenGL at the same time.
    This class holds a weak ref to the SLM object, so that the SLM object can be garbage collected.
    """

    _lock = threading.RLock()

    def __init__(self, slm):
        if isinstance(slm, self.__class__):
            # copy constructor
            self._slm = slm._slm
            if slm.slm is None:
                raise ValueError("The SLM object has been deleted")
        else:
            # construct a new Context object
            # note that adding the same slm to the set multiple times
            # is not a problem, because _all_slms is a set
            self._slm = weakref.ref(slm)

    def __enter__(self):
        self._lock.acquire()
        slm = self.slm
        if slm is not None:
            glfw.make_context_current(slm._window)  # noqa: ok to use _window
        return slm

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._lock.release()

    @property
    def slm(self) -> SLM:
        return self._slm()



import astropy.units as u
import numpy as np

from openwfs.utilities import (
    set_pixel_size,
    get_pixel_size,
    place,
    Transform,
    project,
)


def test_to_matrix():
    # Create a transform object
    transform = Transform(
        transform=((1, 2), (3, 4)),
        source_origin=(0.0, 0.0) * u.m,
        destination_origin=(0.001, 0.002) * u.mm,
    )

    # Define the expected output matrix for same input and output pixel sizes
    expected_matrix = ((1, 2, 1), (3, 4, 1), (0, 0, 1))
    result_matrix = transform.to_matrix((1, 2) * u.um, (1, 2) * u.um)
    assert result_matrix.shape == (3, 3)
    assert np.allclose(result_matrix, expected_matrix)

    # Repeat for different input and output pixel sizes
    expected_matrix = ((0.5, 4, 1), (1.5, 8, 1), (0, 0, 1))
    result_matrix = transform.to_matrix((0.5, 4) * u.um, (1, 2) * u.um)
    assert np.allclose(result_matrix, expected_matrix)

    # Test center correction. The center of the source image should be mapped to the center of the destination image
    src = (17, 18)
    dst = (13, 14)
    src_center = np.array((0.5 * (src[1] - 1), 0.5 * (src[0] - 1), 1.0))
    dst_center = np.array((0.5 * (dst[1] - 1), 0.5 * (dst[0] - 1)))
    transform = Transform()
    result_matrix = transform.cv2_matrix(
        source_shape=src,
        source_pixel_size=(1, 1),
        destination_shape=dst,
        destination_pixel_size=(1, 1),
    )
    assert np.allclose(result_matrix @ src_center, dst_center)

    # Test center correction. The center of the source image should be mapped to the center of the destination image
    transform = Transform()  # transform=((1, 2), (3, 4)))
    result_matrix = transform.cv2_matrix(
        source_shape=src,
        source_pixel_size=(0.5, 4) * u.um,
        destination_shape=dst,
        destination_pixel_size=(1, 2) * u.um,
    )
    assert np.allclose(result_matrix @ src_center, dst_center)

    # Also check openGL matrix (has y-axis flipped and extra row and column)
    expected_matrix = ((1, 2, 1), (3, 4, 2), (0, 0, 1))
    transform = Transform(transform=((1, 2), (3, 4)), source_origin=(0, 0), destination_origin=(1, 2))
    result_matrix = transform.to_matrix((1, 1), (1, 1))
    assert np.allclose(result_matrix, expected_matrix)

    result_matrix = transform.opencl_matrix()
    expected_matrix = ((4, -2, 0, 0), (3, -1, 0, 0), (2, -1, 1, 0))
    assert np.allclose(result_matrix, expected_matrix)


def test_place():
    ps1 = (0.5, 2) * u.um
    src = set_pixel_size(np.random.uniform(size=(7, 8)), ps1)

    # place image in output with same extent and same pixel size
    assert np.all(get_pixel_size(src) == ps1)
    dst_trivial = place(src.shape, ps1, src)
    assert np.allclose(src, dst_trivial)
    assert np.all(get_pixel_size(dst_trivial) == ps1)

    # place image in output with larger extent and same pixel size.
    # the source will be padded with zeros (exactly one row and two columns on all sides)
    dst_pad = place((9, 12), ps1, src)
    assert np.all(dst_pad[0, :] == 0.0)
    assert np.all(dst_pad[-1, :] == 0.0)
    assert np.all(dst_pad[:, -2:] == 0.0)
    assert np.all(dst_pad[:, 0:2] == 0.0)

    # place the image back in an array of the original shape.
    # this should remove the padding
    dst_crop = place((7, 8), ps1, dst_pad)
    assert np.allclose(dst_crop, src)

    # use offset to shift the image
    dst_shift = place((7, 8), ps1, src, offset=ps1 * (2, 1))
    assert np.allclose(dst_shift[0:2, :], 0.0)
    assert np.allclose(dst_shift[:, 0], 0.0)
    assert np.allclose(dst_shift[2:, 1:], src[:-2, :-1])

    # place image in output with same extent and different pixel size
    ps2 = (0.25, 1.0) * u.um
    dst_resample = place((14, 16), ps2, src)
    src_resample = np.repeat(np.repeat(src, 2, axis=0), 2, axis=1)
    assert np.allclose(dst_resample, src_resample)


def test_transform():
    ps1 = (0.5, 2) * u.um
    src = set_pixel_size(np.random.uniform(size=(7, 8)), ps1)

    identity = Transform()
    matrix = identity.cv2_matrix(src.shape, ps1, src.shape, ps1)
    assert np.allclose(matrix, ((1.0, 0.0, 0.0), (0.0, 1.0, 0.0)))

    # shift both origins by same distance
    t0 = Transform(source_origin=-ps1 * (1.7, 2.2), destination_origin=-ps1 * (1.7, 2.2))
    dst0 = project(
        src,
        source_extent=ps1 * np.array(src.shape),
        transform=t0,
        out_extent=ps1 * np.array(src.shape),
        out_shape=src.shape,
    )
    assert np.allclose(dst0, src)

    # shift source by (1,2) pixel
    t1 = Transform(source_origin=-ps1 * (1, 2))
    dst1a = place(src.shape, ps1, src, offset=ps1 * (1, 2))
    dst1b = project(
        src,
        source_extent=ps1 * np.array(src.shape),
        transform=t1,
        out_extent=ps1 * np.array(src.shape),
        out_shape=src.shape,
    )
    assert np.allclose(dst1a, dst1b)


def test_zoom():
    transform = Transform.zoom(2.0)
    vector = (0.3, 0.4)
    result = transform.apply(vector)
    expected_result = (0.6, 0.8)
    assert np.allclose(result, expected_result)


def test_inverse():
    transform = Transform(
        transform=((0.1, 0.2), (-0.25, 0.33)),
        source_origin=(0.12, 0.15),
        destination_origin=(0.23, 0.33),
    )
    vector = (0.3, 0.4)
    result = transform.apply(vector)

    inverse = transform.inverse()
    transformed_back = inverse.apply(result)
    assert np.allclose(vector, transformed_back)
    assert np.allclose(inverse @ transform @ vector, vector)
    assert np.allclose(transform @ inverse @ vector, vector)
